#!/usr/bin/env python3
#
# Creates a synchronous message Client or Server, and uses it for a message console.
# Reads are JSON, writes are CSV.
#
import os
import sys
import cmd
import traceback
import time
from collections import OrderedDict

from .server import Server
from .client import Client

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)
from msgtools.lib.messaging import Messaging

import msgtools.lib.msgcsv as msgcsv
import msgtools.lib.msgjson as msgjson

class MsgCmd(cmd.Cmd):
    intro = 'Type help or ? to list commands.\n'
    prompt = 'msg> '
    def __init__(self, connection, timeout):
        super(MsgCmd, self).__init__()
        self._connection = connection
        self._timeout = timeout
    
    def can_exit(self):
        return True
    def do_exit(self, s):
        print("^D")
        return True
    do_EOF = do_exit

    def do_send(self, line):
        # this translates the input command from CSV to a message, and sends it.
        msg = msgcsv.csvToMsg(line)
        if msg:
            self._connection.send(msg)
            print("sent " + msgcsv.toCsv(msg))
        else:
            print("ERROR! Invalid msg [%s]!" % (line))
    
    def do_recv(self, line):
        msgIDs = []
        msgIDNames = line.split()
        keep_looping = False
        for msgname in msgIDNames:
            if msgname == 'ALL':
                keep_looping = True
                continue
            try:
                if int(msgname, 0):
                    msgIDs.append(int(msgname,0))
            except ValueError:
                if msgname in Messaging.MsgIDFromName:
                    msgIDs.append(int(Messaging.MsgIDFromName[msgname], 0))
                else:
                    print("invalid msg " + msgname)
        if msgIDs:
            msgNames = [Messaging.MsgNameFromID[hex(int(id))] for id in msgIDs]
            print("recv " + str(msgNames))
        else:
            print("recv ANY")
        # this blocks until message received, or timeout occurs
        timeout = self._timeout # value in seconds
        while True:
            try:
                t1 = time.time()
                msg = self._connection.recv(msgIDs, timeout)
                t2 = time.time()
                if timeout != 0.0:
                    delta_t = t2 - t1
                    timeout = timeout - delta_t
                    if timeout < 0:
                        break
            except KeyboardInterrupt:
                msg = None
                print('')
                return
            if msg:
                # print as JSON for debug purposes
                json = msgjson.toJson(msg)
                print(json)
            else:
                print("{}")
            if not keep_looping:
                break
    
    def do_timeout(self, line):
        self._timeout = float(line)
    
    def do_info(self, line):
        try:
            autocomplete, help = msgcsv.csvHelp(line, verbose=True)
            if autocomplete == line or autocomplete == line+",":
                autocomplete, help = msgcsv.csvHelp(line+",", verbose=True)
            if autocomplete and autocomplete.strip() != line.strip():
                print("Invalid message name: " + line)
            elif help:
                print("")
                print(help)
        except:
            print(traceback.format_exc())
        
    def autocomplete(self, commandName, text, line):
        try:
            # get the text with whitespace collapsed, after first word
            simplified = " ".join(line.split()[1:])
            if line.endswith(' ') and len(line.strip()) > len(commandName + " "):
                simplified = simplified + ' '

            autocomplete, help = msgcsv.csvHelp(simplified, verbose= commandName=="info")
            if autocomplete and autocomplete.strip() != simplified.strip():
                return [autocomplete]
            elif help:
                print("")
                print(help)
                print("\n%s%s" % (self.prompt, line), end='')
        except:
            print(traceback.format_exc())
        return []
        
    def complete_send(self, text, line, start_index, end_index):
        return self.autocomplete("send", text, line)

    def complete_recv(self, text, line, start_index, end_index):
        return self.autocomplete("recv", text, line)

    def complete_info(self, text, line, start_index, end_index):
        return self.autocomplete("info", text, line)

def main(args=None):
    Messaging.LoadAllMessages()

    timeout = 10.0
    if len(sys.argv) > 1 and sys.argv[1] == "server":
        connection = Server(timeout)
    else:
        connection = Client("CLI", timeout)
        
    msg_cmd = MsgCmd(connection, timeout)
    try:
        msg_cmd.cmdloop()
    except KeyboardInterrupt:
        print("^C")
        

if __name__ == "__main__":
    main()



import msgtools.parser.parser as MsgParser
from msgtools.parser.MsgUtils import *

def paramType(field):
    typeStr = field["Type"]
    # need to capitalize first 1-2 letters
    if "int" in typeStr:
        if "Offset" in field or "Scale" in field:
            return typeForScaledInt(field)
        typeStr = typeStr.replace("u", "U")
        typeStr = typeStr.replace("i", "I")
        return typeStr
    if typeStr == "float32":
        return "Float";
    if typeStr == "float64":
        return "Double";
    return "?"

def fieldType(field):
    typeStr = field["Type"]
    # need to capitalize first 1-2 letters
    if "int" in typeStr:
        typeStr = typeStr.replace("u", "U")
        typeStr = typeStr.replace("i", "I")
        return typeStr
    if typeStr == "float32":
        return "Float";
    if typeStr == "float64":
        return "Double";
    return "?"

def bitParamType(field, bitfield):
    ret = paramType(field)
    if "Offset" in bitfield or "Scale" in bitfield:
        return typeForScaledInt(field)
    return ret

def fnHdr(field):
    ret = "// %s %s, (%s to %s)" % (MsgParser.fieldDescription(field), MsgParser.fieldUnits(field), MsgParser.fieldMin(field), MsgParser.fieldMax(field))
    return ret

def typeForScaledInt(field):
    return "Double"

def enumLookup(field):
    lookup  = "if("+ str(field["Enum"])+"Enum.keys.contains(value))\n"
    lookup += "    {\n"
    lookup += "        value = "+ str(field["Enum"])+"Enum[value];\n"
    lookup += "    }\n"
    lookup += "    "
    return lookup

def reverseEnumLookup(field):
    lookup = "if(!enumAsInt)\n"
    lookup += "{\n"
    lookup += "        if(Reverse"+ str(field["Enum"])+"Enum.keys.contains(value))\n"
    lookup += "        {\n"
    lookup += "            value = Reverse" + str(field["Enum"]) + "Enum[value];\n    "
    lookup += "        }\n"
    lookup += "}\n"
    return lookup

def getFn(field):
    loc = str(MsgParser.fieldLocation(field))
    param = ""
    if MsgParser.fieldCount(field) > 1:
        loc += "+idx*" + str(MsgParser.fieldArrayElementOffset(field))
        param += "_ idx: Int"
    ret = '''\
%s
public func Get%s(%s) -> %s
{
''' % (fnHdr(field), field["Name"], param, paramType(field))
    access = "m_data.GetField(offset: %s)" % (loc)
    if ("Offset" in field or "Scale" in field):
        ret += '    let valI : '+fieldType(field)+' = '+access+';\n'
        access = getMath("valI", field, "Double")
        ret += '    let valD = '+access+';\n'
        ret += '    return valD;\n'
    else:
        ret += '    return %s;\n' % (access)
    ret += '};'
    if MsgParser.fieldUnits(field) == "ASCII" and (field["Type"] == "uint8" or field["Type"] == "int8"):
        ret += '''
%s
public func Get%sString() -> String
{
    var value = "";
    for i in 0 ..< min(%s, hdr.GetDataLength()-%s)
    {
        let nextChar = Get%s(Int(i));
        if(nextChar == 0)
        {
            break;
        }
        value += String(nextChar);
    }
    return value;
};''' % (fnHdr(field), field["Name"], str(MsgParser.fieldCount(field)), str(MsgParser.fieldLocation(field)), field["Name"])
    return ret

def setFn(field):
    valueString = setMath("value", field, fieldType(field))
    param = "_ value: " + paramType(field)
    loc = str(MsgParser.fieldLocation(field))
    if MsgParser.fieldCount(field) > 1:
        loc += "+idx*" + str(MsgParser.fieldArrayElementOffset(field))
        param += ", _ idx: Int"
    ret = '''\
%s
public func Set%s(%s)
{
    m_data.SetField(offset: %s, value: %s);
};''' % (fnHdr(field), field["Name"], param, loc, valueString)
    if MsgParser.fieldUnits(field) == "ASCII" and (field["Type"] == "uint8" or field["Type"] == "int8"):
        ret += '''
%s
public func Set%sString(value: String)
{
    let stringArray = Array(value.utf8);
    for i in 0 ..< min(%s, stringArray.count)
    {
        Set%s(stringArray[i], i);
    }
};''' % (fnHdr(field), field["Name"], str(MsgParser.fieldCount(field)), field["Name"])
    return ret

def getBitsFn(field, bits, bitOffset, numBits):
    access = "(Get%s() >> %s) & %s" % (field["Name"], str(bitOffset), MsgParser.Mask(numBits))
    param = ""
    ret = '''\
%s
public func Get%s(%s) -> %s
{
''' % (fnHdr(bits), MsgParser.BitfieldName(field, bits), param, bitParamType(field, bits))
    if ("Offset" in bits or "Scale" in bits):
        ret += '    let valI = '+access+';\n'
        access = getMath("valI", bits, "Double")
        ret += '    let valD = '+access+';\n'
        ret += '    return valD;\n'
    else:
        ret += '    return %s;\n' % (access)

    ret += '};'
    return ret

def setBitsFn(field, bits, bitOffset, numBits):
    param = "_ value: " + bitParamType(field, bits)
    ret = '''\
%s
public func Set%s(%s)
{
''' % (fnHdr(bits), MsgParser.BitfieldName(field, bits), param)

    valueString = setMath("value", bits, fieldType(field))
    ret += '''\
    var valI = Get%s(); // read
    valI = valI & ~(%s << %s); // clear our bits
    valI = valI | ((%s & %s) << %s); // set our bits
    Set%s(valI); // write
''' % (field["Name"], MsgParser.Mask(numBits), str(bitOffset), valueString, MsgParser.Mask(numBits), str(bitOffset), field["Name"])
    ret += '};'
    return ret

def accessors(msg):
    gets = []
    sets = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            gets.append(getFn(field))
            sets.append(setFn(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    gets.append(getBitsFn(field, bits, bitOffset, numBits))
                    sets.append(setBitsFn(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return gets+sets

def initField(field):
    if "Default" in field:
        if MsgParser.fieldCount(field) > 1:
            ret = "for i in 0 ..< "+str(MsgParser.fieldCount(field))+"\n"
            ret += "{\n"
            ret += "    Set" + field["Name"] + "(" + str(field["Default"]) + ", i);\n"
            ret += "}\n"
            return ret;
        else:
            return  "Set" + field["Name"] + "(" + str(field["Default"]) + ");"
    return ""

def initBitfield(field, bits):
    if "Default" in bits:
        return  "Set" + MsgParser.BitfieldName(field, bits) + "(" +str(bits["Default"]) + ");"
    return ""

def initCode(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            fieldInit = initField(field)
            if fieldInit:
                ret.append(fieldInit)
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    bits = initBitfield(field, bits)
                    if bits:
                        ret.append(bits)

    return ret

def enums(e):
    ret = ""
    for enum in e:
        if enum["Name"] == "IDs":
            for option in enum["Options"]:
                ret += "static let MSG_ID_" + option["Name"] + " = "+str(option["Value"])+";\n"
        else:
            # enum
            fwd = "public enum " + enum["Name"]+"Values: Int {\n"
            for option in enum["Options"]:
                fwd += "    case "+str(option["Name"]) + " = "+str(option["Value"])+";\n"
            fwd += "}\n"

            # forward map
            fwd += "public static let "+enum["Name"]+"Enum = ["
            for option in enum["Options"]:
                fwd += '"'+str(option["Name"])+'"' +': '+str(option["Value"]) + ', '
            fwd = fwd[:-2]
            fwd += "]\n"
            
            # Reverse map
            back = "public static let Reverse" + enum["Name"]+"Enum = ["
            for option in enum["Options"]:
                back += str(option["Value"]) +': "'+str(option["Name"]) + '", '
            back = back[:-2]
            back += "]\n"

            ret += fwd+back
    return ret

def reflectionInterfaceType(field):
    type = field["Type"]
    if "float" in type or "Offset" in field or "Scale" in field:
        type = "float"
    elif MsgParser.fieldUnits(field) == "ASCII":
        type = "string"
    elif "Enum" in field:
        type = "enumeration"
    else:
        type = "int"
    return type

def bitsReflectionInterfaceType(field):
    type = "int"
    if "Offset" in field or "Scale" in field:
        type = "float"
    elif MsgParser.fieldUnits(field) == "ASCII":
        type = "string"
    elif "Enum" in field:
        type = "enumeration"
    else:
        type = "int"
    return type

def bitfieldReflection(msg, field, bits):
    name = bits["Name"]
    ret = "["+\
              '"name":"'+name + '",'+\
              '"type":"'+bitsReflectionInterfaceType(bits) + '",'+\
              '"units":"'+MsgParser.fieldUnits(bits) + '",'+\
              '"minVal":"'+str(MsgParser.fieldMin(bits)) + '",'+\
              '"maxVal":"'+str(MsgParser.fieldMax(bits)) + '",'+\
              '"description":"'+MsgParser.fieldDescription(bits) + '",'+\
              '"get":"Get' + name + '",'+\
              '"set":"Set' + name  + '", '
    if "Enum" in bits:
        ret += '"enumLookup" : ['+  bits["Enum"]+"Enum, " + "Reverse" + bits["Enum"]+"Enum]]"
    else:
        ret += '"enumLookup" : []]'
    return ret

def fieldReflection(msg, field):
    fieldFnName = field["Name"]
    fieldCount = MsgParser.fieldCount(field)
    if MsgParser.fieldCount(field) != 1 and MsgParser.fieldUnits(field) == "ASCII" and (field["Type"] == "uint8" or field["Type"] == "int8"):
        fieldFnName = field["Name"]+"String"
        fieldCount = 1
    fieldInfo = "["+\
                  '"name":"'+field["Name"] + '",'+\
                  '"type":"'+reflectionInterfaceType(field) + '",'+\
                  '"units":"'+MsgParser.fieldUnits(field) + '",'+\
                  '"minVal":"'+str(MsgParser.fieldMin(field)) + '",'+\
                  '"maxVal":"'+str(MsgParser.fieldMax(field)) + '",'+\
                  '"description":"'+MsgParser.fieldDescription(field) + '",'+\
                  '"get":"Get' + fieldFnName + '",'+\
                  '"set":"Set' + fieldFnName  + '",'+\
                  '"count":'+str(fieldCount) + ', '
    if "Bitfields" in field:
        bitfieldInfo = []
        for bits in field["Bitfields"]:
            bitfieldInfo.append("    " + bitfieldReflection(msg, field, bits))
        fieldInfo += '"bitfieldInfo" : [\n' + ",\n".join(bitfieldInfo) + "], "
    else:
        fieldInfo += '"bitfieldInfo" : [], '
    if "Enum" in field:
        fieldInfo += '"enumLookup" : [' + field["Enum"]+"Enum, " + "Reverse" + field["Enum"]+"Enum]]"
    else:
        fieldInfo += '"enumLookup" : []]'
    return fieldInfo

def reflection(msg):
    fieldInfos = []
    if "Fields" in msg:
        for field in msg["Fields"]:
            fieldInfos.append(fieldReflection(msg, field))
    return ",\n".join(fieldInfos)

def genericInfo(field, loc, type):
    params  = '    static final int loc   = ' + loc + ';\n'
    params += '    static final '+type+' max   = ' + fieldMax(field) + ';\n'
    params += '    static final '+type+' min   = ' + fieldMin(field) + ';\n'
    params += '    static final String units = "' + str(MsgParser.fieldUnits(field)) + '"' + ';\n'
    params += '    static final int count = ' + str(MsgParser.fieldCount(field)) + ';\n'
    if "Default" in field:
        params += '    static final '+type+' defaultValue = ' + str(field["Default"]) + ";\n" 
    if "Scale" in field:
        params += '    static final int scale = ' + str(field["Scale"]) + ';\n'
    if "Offset" in field:
        params += '    static final int offset = ' + str(field["Offset"]) + ';\n'
    return params
    
def fieldInfo(field):
    retType = fieldType(field)
    if "Offset" in field or "Scale" in field:
        retType = typeForScaledInt(field)

    params  = 'class ' + field["Name"] + 'FieldInfo {\n'
    params += genericInfo(field, str(MsgParser.fieldLocation(field)), retType)
    params += '};\n'
    return params

def fieldBitsInfo(field, bits, bitOffset, numBits):
    retType = fieldType(field)
    if "Offset" in bits or "Scale" in bits:
        retType = typeForScaledInt(bits)

    params  = 'class ' + bits["Name"] + 'FieldInfo {\n'
    params += genericInfo(bits, str(MsgParser.fieldLocation(field)), retType)
    params += '    static final int bitOffset = ' + str(bitOffset) + ';\n'
    params += '    static final int numBits   = ' + str(numBits) + ';\n'
    params += '};\n'
    return params

def fieldInfos(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            ret.append(fieldInfo(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    ret.append(fieldBitsInfo(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return "\n".join(ret)
    
def declarations(msg, msg_enums):
    return []

def getMsgID(msg):
    prefix = "var ret : UInt32 = 0;\n"
    mods = ""
    if "Fields" in msg:
        for field in msg["Fields"]:
            if "IDBits" in field:
                numBits = str(field["IDBits"])
                getStr = "UInt32(Get"+field["Name"]+"())"
                if mods:
                    mods += "ret = ret << " + numBits + ";\n"
                mods += "ret += " + getStr + ";\n"
            if "Bitfields" in field:
                for bitfield in field["Bitfields"]:
                    if "IDBits" in bitfield:
                        numBits = str(bitfield["IDBits"])
                        getStr = "UInt32(Get"+BitfieldName(field, bitfield)+"())"
                        if mods:
                            mods += "ret = ret << " + numBits + ";\n"
                        mods += "ret += " + getStr + ";\n"
    return prefix + mods + "return ret;"
    
def setMsgID(msg):
    ret = ""
    numBits = 0
    if "Fields" in msg:
        for field in reversed(msg["Fields"]):
            if "IDBits" in field:
                if numBits != 0:
                    ret += "\nret = ret >> " + str(numBits)+"\n"
                numBits = field["IDBits"]
                setStr = "ret & "+Mask(numBits)
                setStr = fieldType(field)+"("+setStr+")"
                ret +=  "Set"+field["Name"]+"("+setStr+")"
            if "Bitfields" in field:
                for bitfield in reversed(field["Bitfields"]):
                    if "IDBits" in bitfield:
                        if numBits != 0:
                            ret += "\nid = ret >> " + str(numBits)+"\n"
                        numBits = bitfield["IDBits"]
                        setStr = "ret & "+Mask(numBits)
                        setStr = fieldType(field)+"("+setStr+")"
                        ret +=  "Set"+bitfield["Name"]+"("+setStr+")"
    if ret.count('\n') > 1:
        ret = "var ret = id\n" + ret
    else:
        ret = "let ret = id\n" + ret
    return ret

def structUnpacking(msg):
    ret = []

    if "Fields" in msg:    
        for field in msg["Fields"]:
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    ret.append('try { ret["'+bits["Name"] + '"] = Get' + bits["Name"] + "(); } catch (err) {}")
            else:
                if MsgParser.fieldCount(field) == 1:
                    ret.append('try { ret["'+field["Name"] + '"] = Get' + field["Name"] + "(); } catch (err) {}")
                else:
                    if MsgParser.fieldUnits(field) == "ASCII" and (field["Type"] == "uint8" or field["Type"] == "int8"):
                        ret.append('try { ret["'+field["Name"] + '"] = Get' + field["Name"] + "String(); } catch (err) {}")
                    else:
                        ret.append('try { ret["'+field["Name"] + '"] = []; } catch (err) {}')
                        ret.append('try { ')
                        ret.append("    for i in 0 ..< "+str(MsgParser.fieldCount(field)))
                        ret.append('    {')
                        ret.append('        ret["'+field["Name"] + '"][i] = Get' + field["Name"] + "(i);")
                        ret.append('    }')
                        ret.append('} catch (err) {}')
            
    return "\n".join(ret)



#!/usr/bin/env python3
#
# Client is a class that uses a TCP client socket to allow synchronous code to
# send and receives messages.  All instances of Client share a single TCP socket,
# and can also communicate amongst themselves whether the socket is connected or
# not.
#
import gevent.queue
import gevent.monkey
gevent.monkey.patch_socket()
import os
import sys
import socket
import time
from msgtools.lib.messaging import Messaging
from msgtools.sim.sim_exec import SimExec

class Client:
    # keep a list of all clients, so that when any of them send anything it goes to the others.
    _clients = []
    # One socket, shared by all clients
    # Each client blocks on reading from it's own gevent.queue
    _sock = None
    _socket_connected = False
    # A totally separate coroutine does a blocking read with infinite timeout,
    # and writes to all the Client queues.
    _rx_greenlet = None
    # A combined name for all clients
    _name = None
    # record particular messages the user wants to record, even if they never called recv on them
    _extra_msgs_to_record = {}
    received = {}
    def __init__(self, name='Client', timeout=10):
        # keep a reference to Messages, for convenience of cmdline scripts
        self.Messages = Messaging.Messages
        # load all messages if not already loaded
        if Messaging.hdr == None:
            Messaging.LoadAllMessages()

        self._timeout = timeout

        # store combined name
        if Client._name == None:
            Client._name = name
        else:
            Client._name = Client._name + "," + name

        if Client._sock == None and SimExec.allow_socket_comms:
            # if there isn't yet a socket but there should be,
            # call the function to open it.
            Client.reconnect_socket(False)
            # Start a greenlet thread to read from the socket.
            # If the socket closes or has an error, the thread will reconnect it.
            Client._rx_greenlet = gevent.spawn(Client.read_for_all_clients)

        # make a queue for receiving messages from other clients in this
        # same process and from the socket.
        self._rx_queue = gevent.queue.Queue()

        # keep a dictionary of the latest value of all received messages
        self.received = {}
        
        # add us to the list of clients
        Client._clients.append(self)

    # This will be called once on startup with retry=False, and after that,
    # will only be called from the read_msg_from_socket() greenlet thread,
    # with retry=True.  Because of that, if retry=True, we'll gevent.sleep()
    # when called with retry==True, to prevent too much network activity.
    @staticmethod
    def reconnect_socket(retry=True):
        if retry:
            # If we're trying to reconnect, wait a little bit so we don't spend
            # a ton of time doing network stuff in the case where the server
            # isn't even running.
            gevent.sleep(0.5)
        try:
            Client._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            Client._sock.connect(("127.0.0.1", 5678))
            Client._sock.setblocking(True)
            Client._sock.settimeout(None)

            # do default subscription to get *everything*
            subscribeMsg = Messaging.Messages.Network.MaskedSubscription()
            Client.static_send(subscribeMsg)

            # Send the connect message with the name
            connectMsg = Messaging.Messages.Network.Connect()
            connectMsg.SetName(Client._name)
            Client.static_send(connectMsg)
            Client._socket_connected = True
        except:
            Client._socket_connected = False
            #print("couldn't open socket!!")
            if not SimExec.sim_exists:
                raise

    @staticmethod
    def static_send(msg):
        if Client._sock == None:
            pass#return

        try:
            bufferSize = len(msg.rawBuffer().raw)
            computedSize = msg.hdr.SIZE + msg.hdr.GetDataLength()
            if(computedSize > bufferSize):
                msg.hdr.SetDataLength(bufferSize - msg.hdr.SIZE)
                print("client.py: Truncating message %s from %d to %d bytes" % (msg.__class__.__name__, computedSize, bufferSize))
            if(computedSize < bufferSize):
                # don't send the *whole* message, just a section of it up to the specified length
                Client._sock.send(msg.rawBuffer().raw[0:computedSize])
            else:
                Client._sock.send(msg.rawBuffer().raw)

        except Exception as e:
            if not SimExec.sim_exists:
                raise
        
    # supposedly it's dangerous for multiple greenlets to send to the same socket!
    #TODO If this causes problems, we'll need to use tx queues to send to a single greenlet,
    #TODO and then have it put data in the socket (like the reverse of how rx works)
    def send(self, msg):
        Client.static_send(msg)
        Client.queue_for_clients(msg, self)

    @staticmethod
    def queue_for_clients(msg, excluded_client):
        if len(Client._clients) > 0:#1:
            for c in Client._clients:
                if c != excluded_client:
                    c._rx_queue.put(msg)

    @staticmethod
    def read_msg_from_socket():
        try:
            # see if there's enough for header
            data = Client._sock.recv(Messaging.hdr.SIZE, socket.MSG_PEEK)
            if len(data) == Messaging.hdr.SIZE:
                # create header based on peek'd data
                hdr = Messaging.hdr(data)

                # see if there's enough for the body, too
                data += Client._sock.recv(hdr.GetDataLength(), socket.MSG_PEEK)
                if len(data) != Messaging.hdr.SIZE + hdr.GetDataLength():
                    #print("didn't get whole body, error!")
                    return None

                # read out what we peek'd.
                data = Client._sock.recv(Messaging.hdr.SIZE + hdr.GetDataLength())

                # reset the header based on appended data
                hdr = Messaging.hdr(data)
                msg = Messaging.MsgFactory(hdr)
                return msg
            else:
                # reopen the socket
                Client.reconnect_socket()
        except ConnectionResetError:
            Client.reconnect_socket()
        except OSError:
            Client.reconnect_socket()
        except socket.timeout:
            print("timeout")
            return None
        except BlockingIOError:
            print('blocking')
            return None
        return None

    @staticmethod
    def read_for_all_clients():
        while True:
            msg = Client.read_msg_from_socket()
            if msg:
                #TODO Should we put everything in the queue?
                #TODO The client might decide to receive a message by ID
                #TODO after it got read from the socket, in which case we'd miss it.
                #TODO That's slightly different than in the single Client version
                #TODO where a Client reads directly from the scoket because then
                #TODO nothing is read from the socket until the Client
                #TODO decides to read and passes a list of msgIds.
                Client.queue_for_clients(msg, None)
            else:
                # If the server disconnects we'll get to this case.
                # That's ok because read_msg_from_socket() will attempt
                # to reconnect.
                pass

    def recv(self, msgIds=[], timeout=None):
        # if user didn't pass a list, put the single param into a list
        if not isinstance(msgIds, list):
            msgIds = [msgIds]
        # if they passed classes, get the ID of each
        for i in range(0,len(msgIds)):
            if hasattr(msgIds[i], 'ID'):
                #print("replacing %s with %s" % (msgIds[i], msgIds[i].ID))
                msgIds[i] = msgIds[i].ID

        # reset the timeout member variable
        if timeout != None and self._timeout != timeout:
            self._timeout = timeout

        if self._timeout == 0.0:
            block = False
            timeout = None
        else:
            block = True
            timeout = self._timeout

        # Process any messages in our rx queue
        while True:
            if block and SimExec.sim_exists:
                # Delay for a period of simulation time, not real time!
                start_time = SimExec.time()
                while True:
                    try:
                        msg = self._rx_queue.get(block=False)
                    except KeyboardInterrupt:
                        raise
                    except:
                        SimExec.tick_event.wait()
                        SimExec.tick_event.clear()
                        if SimExec.time() > start_time + timeout:
                            return None
            else:
                # if we dont't want to block, or else the sim isn't running,
                # then we can read from the queue directly without
                # worrying about simulation time.
                try:
                    msg = self._rx_queue.get(block, timeout)
                except KeyboardInterrupt:
                    raise
                except:
                    return None

            id = msg.hdr.GetMessageID()
            if len(msgIds) == 0 or id in msgIds:
                self.received[id] = msg
                return msg

    def wait_for(self, msgIds=[], fn=None, timeout=None):
        ret = False
        if timeout == None:
            timeout = self._timeout
        remaining_timeout = timeout
        while(1):
            before = SimExec.time()
            msg = self.recv(msgIds, remaining_timeout)
            if msg != None and fn(msg):
                ret = True
                break
            after = SimExec.time()
            elapsed = after - before
            if remaining_timeout:
                # if we have a timeout, update it for the amount of time spent
                remaining_timeout = remaining_timeout - elapsed
                if remaining_timeout < 0:
                    ret = False
                    break
        return ret

    def stop(self):
        Client._clients.remove(self)

    # select set of messages to record the last value of
    def record(self, msgs_to_record):
        Client._extra_msgs_to_record.update(msgs_to_record)



import sys, os, multiprocessing
from PyQt5 import QtWidgets, QtCore, QtGui

from msgtools.script.edit_window import EditWindow
from msgtools.script import debugger as debugger
import msgtools.script.launcher as launcher

NEW_FILE = '''\
from msgtools.lib.messaging import Messaging as M
from msgtools.lib.message import Message as Msg
from msgtools.console.client import Client

M.LoadAllMessages()
cxn = Client('example')
'''

class MsgScript(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        super(MsgScript, self).__init__(parent)
        self.setWindowIcon(QtGui.QIcon(launcher.info().icon_filename))
        
        self.settings = QtCore.QSettings("MsgTools", 'MsgScript')
        self.restoreGeometry(self.settings.value("geometry", QtCore.QByteArray()))
        self.restoreState(self.settings.value("windowState", QtCore.QByteArray()))
        
        # status bar
        self.statusBar().showMessage("")
                
        # menu bar
        newAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("document-new"), '&New', self)
        openAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("document-open"), '&Open', self)
        closeAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("document-close"), '&Close', self)
        saveAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("document-save"), '&Save', self)
        saveAsAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("document-save-as"), 'Save &As', self)
        menubar = self.menuBar()
        file_menu = menubar.addMenu('&File')
        file_menu.addAction(newAction)
        file_menu.addAction(openAction)
        file_menu.addAction(closeAction)
        file_menu.addAction(saveAction)
        file_menu.addAction(saveAsAction)
        newAction.triggered.connect(self.new_action)
        openAction.triggered.connect(self.open_action)
        closeAction.triggered.connect(self.close_action)
        saveAction.triggered.connect(self.save_action)
        saveAsAction.triggered.connect(self.save_as_action)

        pauseAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("media-playback-pause"), "&Pause", self)
        stepAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("media-skip-forward"), "Step &Into", self)
        stepOutAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("media-seek-forward"), "Step &Oout", self)
        runAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("media-playback-start"), "&Run", self)
        stopAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("media-playback-stop"), "&Stop", self)
        clearAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("user-trash"), "&Clear", self)
        stepTimeSlider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        stepTimeSlider.setMinimum(0)
        stepTimeSlider.setMaximum(3)
        stepTimeSlider.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)

        debug_menu = menubar.addMenu('&Debug')
        debug_menu.addAction(runAction)
        debug_menu.addAction(pauseAction)
        debug_menu.addAction(stopAction)
        debug_menu.addAction(clearAction)
        runAction.triggered.connect(self.run_action)
        stepAction.triggered.connect(self.step_action)
        stepOutAction.triggered.connect(self.step_out_action)
        pauseAction.triggered.connect(self.pause_action)
        stopAction.triggered.connect(self.stop_action)
        clearAction.triggered.connect(self.clear_action)
        
        # toolbars
        file_toolbar = self.addToolBar("File")
        file_toolbar.setObjectName("file_toolbar")
        file_toolbar.addAction(newAction)
        file_toolbar.addAction(openAction)
        file_toolbar.addAction(closeAction)
        file_toolbar.addAction(saveAction)
        file_toolbar.addAction(saveAsAction)
        
        debug_toolbar = self.addToolBar("Debug")
        debug_toolbar.setObjectName("debug_toolbar")
        debug_toolbar.addAction(runAction)
        debug_toolbar.addAction(stepAction)
        debug_toolbar.addAction(stepOutAction)
        debug_toolbar.addAction(pauseAction)
        debug_toolbar.addAction(stopAction)
        debug_toolbar.addAction(clearAction)
        debug_toolbar.addWidget(QtWidgets.QLabel("Step Time"))
        debug_toolbar.addWidget(stepTimeSlider)
        
        self.setWindowTitle("MsgScript")

        # editor/debug window
        self.editor = EditWindow()
        self.editor.modificationChanged.connect(self.document_was_modified)
        self.editor.setText("")
        self.current_filename = ""
        last_filename = self.settings.value("last_filename", '')
        if last_filename:
            self.load_file(last_filename)
            
        # script output window
        self.scriptOutput = QtWidgets.QPlainTextEdit(self)
        
        self.splitter = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        self.splitter.addWidget(self.editor)
        self.splitter.addWidget(self.scriptOutput)
        self.setCentralWidget(self.splitter)
        self.splitter.restoreState(self.settings.value("SplitterSize", self.splitter.saveState()));
        
        # Initialize the debug and application queues for passing messages
        self.debugq = multiprocessing.Queue()
        self.applicationq = multiprocessing.Queue()
        
        self.debugprocess = None
        self.isrunning = False
        
        self.debug_timer = QtCore.QTimer(self)
        self.debug_timer.setSingleShot(False)
        self.debug_timer.timeout.connect(self.poll_output)
        # this determines how quickly we notice the debugger did something,
        # and in turn limits the rate of 'isrunning' auto-stepping
        stepTimeSlider.setValue(int(self.settings.value("step_time", 1)))
        self.time_slider_changed(stepTimeSlider.value())
        stepTimeSlider.valueChanged.connect(self.time_slider_changed)

    # poll output from the applicationq, which is another Process
    def poll_output(self):
        while not self.applicationq.empty():
            appinfo = self.applicationq.get()
            if 'exception' in appinfo:
                self.write("Exception: " + str(appinfo['exception']))
            elif 'stdout' in appinfo:
                self.write(str(appinfo['stdout']))
            elif 'stderr' in appinfo:
                self.write(str(appinfo['stderr']))
            elif 'error' in appinfo:
                self.editor.crashed()
            elif 'exit' in appinfo:
                self.stop_action()
            elif 'trace' in appinfo:
                tr = appinfo['trace']
                co = appinfo['co']
                #self.write("%s: line %d\n" % (co['file'], co['lineno']))
                if co['file'] == self.current_filename:
                    if self.debugprocess and self.debugprocess.is_alive():
                        self.editor.ran_to_line(co['lineno'])
                        if self.isrunning:
                            if not self.editor.has_breakpoint(co['lineno']):
                                self.debugq.put('step')
                else:
                    #open another file, switch to it's tab, show the line we're at!?!?
                    pass

    def write(self, message):
        self.scriptOutput.moveCursor(QtGui.QTextCursor.End)
        self.scriptOutput.insertPlainText(message)
        self.scriptOutput.moveCursor(QtGui.QTextCursor.End)

    def new_action(self):
        if self.maybe_save():
            self.set_current_file('')
            self.editor.setText(NEW_FILE)
            self.editor.setModified(False)
            self.set_window_modified()
    
    def open_action(self):
        if self.maybe_save():
            filename, _ = QtWidgets.QFileDialog.getOpenFileName(self)
            if filename:
                self.load_file(filename)
    
    def save_action(self):
        if self.current_filename == '':
            return self.save_as_action()
        else:
            return self.save_file(self.current_filename)
    
    def save_as_action(self):
        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.AcceptSave
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(self,"QtWidgets.QFileDialog.getSaveFileName()","","All Files (*);;Text Files (*.txt)", options=options)
        if not filename:
            return False
        return self.save_file(filename)
    
    def save_file(self, filename):
        try:
            file = open(filename, 'w')
        except IOError:
            QtWidgets.QMessageBox.warning(self, "Application", "Cannot write file %1" % (filename))
            return
        else:
            with file:
                file.write(self.editor.text())
                self.set_current_file(filename)
                self.statusBar().showMessage("File saved", 2000)
        return True
    
    def load_file(self, filename):
        try:
            file = open(filename, 'r')
        except (IOError, FileNotFoundError):
            QtWidgets.QMessageBox.warning(self, "Application", "Cannot read file %s" % (filename))
            return
        else:
            with file:
                self.set_current_file(filename)
                self.editor.setText(file.read())
                self.editor.setModified(False)
                self.set_window_modified()

    def set_current_file(self, filename):
        self.editor.setModified(False)
        self.current_filename = filename
        self.set_window_modified()

    def set_window_modified(self):
        title = "MsgScript - %s" % self.current_filename
        if self.editor.isModified():
            title = title + " *"
        self.setWindowTitle(title)
    
    def document_was_modified(self, modified):
        if modified:
            self.set_window_modified()
        
    def maybe_save(self):
        if not self.editor.isModified():
            return True
        ret = QtWidgets.QMessageBox.warning(self, "Application",
                                   "The document has been modified.\nDo you want to save your changes?",
                                   QtWidgets.QMessageBox.Save | QtWidgets.QMessageBox.Discard | QtWidgets.QMessageBox.Cancel);
        if ret == QtWidgets.QMessageBox.Save:
            return self.save_file(self.current_filename)
        elif ret == QtWidgets.QMessageBox.Cancel:
            return False

        return True

    def close_action(self):
        self.maybe_save()
        self.set_current_file('')
        self.editor.setText('')
        self.editor.setModified(False)
        self.set_window_modified()
    
    def closeEvent(self, ev):
        if self.maybe_save():
            self.stop_action()
            self.settings.setValue("geometry", self.saveGeometry())
            self.settings.setValue("windowState", self.saveState())
            self.settings.setValue("last_filename", self.current_filename)
            self.settings.setValue("SplitterSize", self.splitter.saveState())
            ev.accept()
        else:
            ev.ignore()
    
    #We want to update the display as the script runs.  Therefore, we should never do "self.debugq.put('over')"
    #We should always step through the code, but we should track if we're "running", and if so, have reading a trace cause another 'step'
    def run_action(self):
        if self.debugprocess != None:
            if self.debugprocess.is_alive():
                #self.debugq.put('over')
                self.isrunning = True
                if self.debugq.empty():
                    self.debugq.put('step')
            else:
                self.debugprocess = None
        if self.debugprocess == None:
            # save the file
            self.save_file(self.current_filename)
            # clear the queues, in case there was anything left over from last run
            self.debugq = multiprocessing.Queue()
            self.applicationq = multiprocessing.Queue()
            # Create the debug process
            self.debugprocess = multiprocessing.Process(target=debugger.debug, args=(self.applicationq, self.debugq, self.current_filename))
            self.debugprocess.start()
            #self.debugprocess.exited.connect(self.process_ended)
    
    def step_action(self):
        if self.debugprocess:
            if self.debugq.empty():
                self.debugq.put('step')
        else:
            self.write('not running')
    
    def step_out_action(self):
        if self.debugprocess:
            if self.debugq.empty():
                self.debugq.put('over')
        else:
            self.write('not running')
    
    def pause_action(self):
        self.isrunning = False
        pass
    
    def stop_action(self):
        self.isrunning = False
        if self.debugprocess != None:
            i = 0
            while(self.debugprocess.is_alive()):
                if i > 0:
                    self.write("Tried to terminate and join(0.1), but still is_alive()!")
                i+=1
                if i > 5:
                    self.write("Couldn't kill script!")
                    return
                self.debugprocess.terminate()
                # This can hang forever!  For some reason the join may block forever, but if
                # we let it timeout is_alive() will return false and we break out of the loop!
                self.debugprocess.join(0.1)
            self.debugprocess = None
            self.editor.exited()
        else:
            self.editor.ran_to_line(0)
    
    def clear_action(self):
        self.scriptOutput.clear()
    
    def time_slider_changed(self, value):
        self.settings.setValue("step_time", value)
        time_val = 100
        if value == 0:
            time_val = 10
        elif value == 1:
            time_val = 100
        elif value == 2:
            time_val = 500
        elif value == 3:
            time_val = 1000
        self.debug_timer.start(time_val)

def main():
    # quiet icon warnings
    os.environ["QT_LOGGING_RULES"] = "qt.svg.warning=false"
    app = QtWidgets.QApplication(sys.argv)
    msgApp = MsgScript()
    msgApp.show()
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()



if 1:
    import json
    import os
    from packaging import version
    import pkg_resources
    try:
        msgtools_distribution = pkg_resources.get_distribution("msgtools")
        install_dir = msgtools_distribution.egg_info
        msgtools_version = version.parse(msgtools_distribution.version)

        version_filename = ".msgtools_requirement.json"
        searchdir = os.getcwd()
        while(True):
            required_version_file = os.path.join(searchdir, version_filename)
            if os.path.isfile(required_version_file):
                with open(required_version_file, 'r') as version_file:
                    j = json.load(version_file)
                    required_version = version.parse(j["version"])
                    if msgtools_version < required_version:
                        print("ERROR!")
                        print("msgtools version %s < required version %s" % (msgtools_version, required_version))
                        print("Required version is set by the contents of:\n    %s" % (required_version_file))
                        print("To resolve this, you should update msgtools, perhaps with:")
                        if install_dir.endswith(".egg-info"):
                            print("    cd %s" % (install_dir.replace("msgtools.egg-info", "")))
                            print("    git pull")
                            print("    make undevelop ; make develop")
                        else:
                            print("    pip3 install --upgrade msgtools")
                        exit(1)

            lastsearchdir = searchdir
            searchdir = os.path.abspath(os.path.join(searchdir, os.pardir))
            # checking if joining with pardir returns the same dir is the easiest way to
            # determine if we're at root of filesystem, and our upward search has to stop.
            if lastsearchdir == searchdir:
                break
    except pkg_resources.DistributionNotFound:
        # If msgtools is not installed, but we are somehow running anyway,
        # then don't do version checks.
        pass



from collections import namedtuple
import ctypes
import datetime
import struct

from .messaging import Messaging

def hexbytes(hdr):
    try:
        b = hdr.rawBuffer()[:]
    except:
        b = hdr
    return "0x"+":".join("{:02x}".format(c) for c in b)

def Crc16(data):
    crc = 0;
    for i in range(0,len(data)):
        d = struct.unpack_from('B', data, i)[0]
        crc = (crc >> 8) | (crc << 8)
        crc ^= d
        crc ^= (crc & 0xff) >> 4
        crc ^= crc << 12
        crc = 0xFFFF & crc
        crc ^= (crc & 0xff) << 5
        crc = 0xFFFF & crc
    return crc

# This helps finalize and validate headers that have any of:
# 1) A start sequence, which is a constant string of bytes that delimits
#    the start of a header, useful for sending data across noisy datalinks
#    that can drop bytes.
# 2) A header CRC-16, that validates the contents of a header
# 3) A body CRC-16, that validates the body of a message.
class HeaderHelper:
    def __init__(self, hdr_class, error_fn):
        self._error_fn = error_fn
        self._hdr_class = hdr_class
        start_sequence_field = Messaging.findFieldInfo(hdr_class.fields, "StartSequence")
        if start_sequence_field != None:
            self._start_sequence = int(hdr_class.GetStartSequence.default)
        else:
            self._start_sequence = None
        try:
            self._hdr_crc_region = int(hdr_class.GetHeaderChecksum.offset)
        except AttributeError:
            self._hdr_crc_region = None

    def header_valid(self, hdr):
        if self._start_sequence != None and hdr.GetStartSequence() != self._start_sequence:
            #self._error_fn("  HDR SS %s != %s" % (hex(hdr.GetStartSequence()) , hex(self._start_sequence)))
            return False
        if self._hdr_crc_region != None:
            # Stop computing before we reach header checksum location.
            headerCrc = Crc16(hdr.rawBuffer()[:self._hdr_crc_region])
            receivedHeaderCrc = hdr.GetHeaderChecksum()
            if headerCrc != receivedHeaderCrc:
                self._error_fn("RX ERROR: HEADER CRC %s != %s for %s" % (hex(headerCrc) , hex(receivedHeaderCrc), hexbytes(hdr)))
                return False
        return True

    def body_valid(self, hdr, body):
        if hdr.GetDataLength() > len(body):
            self._error_fn("RX ERROR: BODY LENGTH %d > %d" % (hdr.GetDataLength(), len(body)))
            return False
        if self._hdr_crc_region != None:
            bodyCrc = Crc16(body[:hdr.GetDataLength()])
            receivedBodyCrc = hdr.GetBodyChecksum()
            if receivedBodyCrc != bodyCrc:
                self._error_fn("RX ERROR: BODY CRC %s != %s for %s" % (hex(receivedBodyCrc), hex(bodyCrc), hexbytes(body)))
                return False
        return True

    def msg_valid(self, msg):
        return self.header_valid(msg.hdr) and self.body_valid(msg.hdr, msg.rawBuffer()[Messaging.hdr.SIZE:])

    def finalize(self, msg):
        if self._hdr_crc_region != None:
            # set header and body CRC
            msg.SetHeaderChecksum(Crc16(msg.rawBuffer()[:self._hdr_crc_region]))
            msg.SetBodyChecksum(Crc16(msg.rawBuffer()[self._hdr_class.SIZE:self._hdr_class.SIZE+msg.GetDataLength()]))

class HeaderTranslator:
    def __init__(self, hdr1, hdr2):
        # Make a list of fields in the headers that have matching names.
        self._correspondingFields = []
        for fieldInfo1 in hdr1.fields:
            if len(fieldInfo1.bitfieldInfo) == 0:
                fieldInfo2 = Messaging.findFieldInfo(hdr2.fields, fieldInfo1.name)
                if fieldInfo2 != None:
                    self._correspondingFields.append([fieldInfo1, fieldInfo2])
            else:
                for bitfieldInfo1 in fieldInfo1.bitfieldInfo:
                    fieldInfo2 = Messaging.findFieldInfo(hdr2.fields, bitfieldInfo1.name)
                    if fieldInfo2 != None:
                        self._correspondingFields.append([bitfieldInfo1, fieldInfo2])
        
        HdrInfo = namedtuple('HdrInfo', 'type infoIndex timeField')
        self._hdr1Info = HdrInfo(hdr1, 0, Messaging.findFieldInfo(hdr1.fields, "Time"))
        self._hdr2Info = HdrInfo(hdr2, 1, Messaging.findFieldInfo(hdr2.fields, "Time"))
        self._timestampOffset = 0
        self._lastTimestamp = 0
        self._lastWrapTime = None

    def translateHdrAndBody(self, fromHdr, body):
        toHdr = self.translateHdr(fromHdr)
        if toHdr:
            # copy the body
            for i in range(0,fromHdr.GetDataLength()):
                toHdr.rawBuffer()[toHdr.SIZE+i] = body[i]
        return toHdr


    def translateHdr(self, fromHdr):
        # figure out which direction to translate
        if isinstance(fromHdr, self._hdr1Info.type):
            fromHdrInfo = self._hdr1Info
            toHdrInfo = self._hdr2Info
        elif isinstance(fromHdr, self._hdr2Info.type):
            fromHdrInfo = self._hdr2Info
            toHdrInfo = self._hdr1Info
        else:
            print("ERROR!  type %s is not %s or %s!" % (type(fromHdr), self._hdr1Info.type, self._hdr2Info.type))
            raise TypeError
        
        # allocate the message to translate to
        toBuffer = ctypes.create_string_buffer(toHdrInfo.type.SIZE+fromHdr.GetDataLength())
        toHdr = toHdrInfo.type(toBuffer)
        toHdr.initialize()

        # loop through fields using reflection, and transfer contents from
        # one header to the other
        for pair in self._correspondingFields:
            fromFieldInfo = pair[fromHdrInfo.infoIndex]
            toFieldInfo = pair[toHdrInfo.infoIndex]
            Messaging.set(toHdr, toFieldInfo, Messaging.get(fromHdr, fromFieldInfo))
        
        # if the message ID can't be expressed in the new header, return None,
        # because this message isn't translatable
        if fromHdr.GetMessageID() != toHdr.GetMessageID():
            if Messaging.debug:
                print("message ID 0x" + hex(fromHdr.GetMessageID()) + " translated to 0x" + hex(toHdr.GetMessageID()) + ", throwing away")
            return None
        
        # do special timestamp stuff to convert from relative to absolute time
        if toHdrInfo.timeField != None:
            def set_time(hdr, t):
                # This is a bit ugly, but it's hard to tell if the Time field is a float or
                # an int.  If it's an int and we give it a float, struct.error gets raised.
                try:
                    hdr.SetTime(t)
                except struct.error:
                    hdr.SetTime(int(t))
            if fromHdrInfo.timeField != None:
                def timeReallyBig(tmax):
                    if tmax == 'DBL_MAX':
                        return True
                    if tmax == 'FLT_MAX':
                        return True
                    if int(tmax) >= 2**32:
                        return True
                def timeScaling(fromUnits, toUnits):
                    if fromUnits.lower() == 'ms' and toUnits.lower() == 's':
                        return 0.001
                    if fromUnits.lower() == 's' and toUnits.lower() == 'ms':
                        return 1000.0
                    return 1.0
                time_scale = timeScaling(fromHdrInfo.timeField.units, toHdrInfo.timeField.units)
                # if we're converting from a header with a smaller timestamp to a header
                # with a bigger timestamp, look for wrapping of the input timestamp
                time_could_wrap = False
                if not timeReallyBig(fromHdrInfo.timeField.maxVal):
                    if timeReallyBig(toHdrInfo.timeField.maxVal):
                        time_could_wrap = True
                    else:
                        time_could_wrap = int(fromHdrInfo.timeField.maxVal) < int(toHdrInfo.timeField.maxVal)
                if time_could_wrap:
                    # Detect time rolling
                    thisTimestamp = fromHdr.GetTime() * time_scale
                    thisTime = datetime.datetime.now()
                    # if the new timestamp is less than the old one plus some margin,
                    # do further checks to see if a wrap occurred.
                    if thisTimestamp < self._lastTimestamp - float(fromHdrInfo.timeField.maxVal) * time_scale * 0.1:
                        # If the check above shows the new timestamp is less than the last timestamp,
                        # count it as wrapping only if the current system time is beyond the last time
                        # a wrap occured by more than 0.5 times the time it should take for the
                        # timestamp to wrap again.  The assumption is that if it's not that far beyond
                        # the last time it wrapped, then messages were sent slightly out-of-order.
                        if (self._lastWrapTime == None or
                            thisTime > self._lastWrapTime + datetime.timedelta(seconds=float(fromHdrInfo.timeField.maxVal) * time_scale * 0.5)):
                            self._lastWrapTime = thisTime
                            self._timestampOffset += 1
                    self._lastTimestamp = thisTimestamp
                    # need to handle different size timestamps!
                    set_time(toHdr, self._timestampOffset * float(fromHdrInfo.timeField.maxVal) * time_scale + thisTimestamp)
                else:
                    set_time(toHdr, fromHdr.GetTime()*time_scale)
            else:
                t = datetime.datetime.now().timestamp()
                maxTime = toHdrInfo.timeField.maxVal
                # use time since start of day, if 32-bit or smaller timestamps
                if maxTime != "DBL_MAX" and (maxTime == "FLT_MAX" or float(maxTime) <= 2**32):
                    t = (datetime.datetime.fromtimestamp(t) - datetime.datetime.fromtimestamp(t).replace(hour=0, minute=0, second=0, microsecond=0)).total_seconds()
                if toHdrInfo.timeField.units == "ms":
                    t = t * 1000.0
                if toHdrInfo.timeField.type == "int":
                    t = int(t)
                set_time(toHdr, t)

        return toHdr

    def translate(self, fromHdr):
        toHdr = self.translateHdrAndBody(fromHdr, fromHdr.rawBuffer()[type(fromHdr).SIZE:])
        if toHdr != None:
            self.promoteIdFields(toHdr)
        return toHdr

    # Handle message body ID fields by populating them in message header.
    # This is to handle systems where some message body fields are conditionally
    # used as ID fields.
    def promoteIdFields(self, hdr):
        # if any message body fields are marked as IDs, look for header fields with
        # same name if header field is zero, set it to value of body field.
        id = hdr.GetMessageID()
        keep_going = True
        while(keep_going):
            keep_going = False
            msgClass = Messaging.MsgClass(hdr)
            msg = msgClass(hdr.rawBuffer())
            for fieldInfo in msg.fields:
                if fieldInfo.idbits != 0:
                    hdrFieldInfo = Messaging.findFieldInfo(hdr.fields, fieldInfo.name)
                    if hdrFieldInfo and hdrFieldInfo.get(hdr) == 0:
                        hdrFieldInfo.set(hdr, fieldInfo.get(msg))
                        keep_going = True



#!/usr/bin/env python3
import json
import sys
import os
import argparse

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)

from msgtools.lib.file_reader import MessageFileReader
from msgtools.lib.messaging import Messaging
import msgtools.lib.msgcsv as msgcsv
import msgtools.lib.msgjson as msgjson

DESCRIPTION='''
    Lumberjack creates either a subdirectory with one CSV file per message type received in that directory,
    or a single JSON file if --json is specified.
'''
class Lumberjack(MessageFileReader):
    def __init__(self):
        super(Lumberjack, self).__init__()

        parser = argparse.ArgumentParser(description=DESCRIPTION)
        parser.add_argument('-o', '--outputdir', help='''Specifies the name of the directory to output
            parsed data to.  Required for non-file connectionTypes.''')
        parser.add_argument('logfile', help='''The log file you want to split into CSV.  
            .log extension assumes the log is binary with NetworkHeaders.  A .txt extension assumes the 
            file was created with SerialHeaders.''')
        parser.add_argument('--json', action='store_true', help='''Causes output to go to a single JSON file instead of a directory of CSV files.''')
        parser.add_argument('--serial', action='store_true', help='''Assumes input file contains binary messages with SerialHeaders instead of NetworkHeaders.''')
        self.args = parser.parse_args()

        if self.args.logfile.lower().endswith('.txt'):
            self.args.serial = True

        # If the user specified the output dir use it.  If the user 
        # specified a source file, then figure out the output dir
        # from the source filename
        if self.args.outputdir is not None:
            self.output_name = self.args.outputdir
        else:
            self.output_name = self.args.logfile.replace('.bin', '')
            self.output_name = self.output_name.replace('.log', '')
            self.output_name = self.output_name.replace('.txt', '')
            self.output_name = self.output_name.replace('.TXT', '')

        if self.args.json:
            self.output_name = self.output_name + ".json"
        else:
            if os.path.exists(self.output_name) is False:
                os.makedirs(self.output_name)
        print("Writing output to " + self.output_name + "\n")

        if self.args.json:
            self.output_file = open(self.output_name, 'w')
        else:
            self.output_files = {}

        self.message_count = 0

        header_name = None
        if self.args.serial:
            header_name = "SerialHeader"
        self.read_file(self.args.logfile, header_name)

    def process_message(self, msg):
        self.message_count += 1
        
        id = msg.hdr.GetMessageID()

        if self.args.json:
            output_file = self.output_file
        else:
            # if we write CSV to multiple files, we'd probably look up a hash table for this message id,
            # and open it and write a header
            if(id in self.output_files):
                output_file = self.output_files[id]
            else:
                # create a new file
                output_file = open(self.output_name + "/" + msg.MsgName().replace("/","_") + ".csv", 'w')

                # store a pointer to it, so we can find it next time (instead of creating it again)
                self.output_files[id] = output_file
                
                # add table header, one column for each message field
                tableHeader = msgcsv.csvHeader(msg, nameColumn=False, timeColumn=True) + '\n'
                output_file.write(tableHeader)

        if self.args.json:
            # Get a dictionary of the message
            dict = msg.toDict(includeHeader=True)

            # the dict has one key, which is the name of the message
            msgname = list(dict.keys())[0]

            text = json.dumps(dict) + '\n'
            output_file.write(text)
        else:
            timestamp = msg.hdr.GetTime()
            text = str(timestamp) + ", "
            text += msgcsv.toCsv(msg, nameColumn=False, timeColumn=False)
            text += '\n'
            output_file.write(text)

def main():
    lumberjack = Lumberjack()

    print("Processed %d messages" % (lumberjack.message_count))

# main starts here
if __name__ == '__main__':
    main()



import os
from PyQt5 import QtCore, QtWidgets
from .socketcan_commands import SocketcanCommands

class SelectCanPortDialog(QtWidgets.QDialog):
    statusUpdate = QtCore.pyqtSignal(str)
    portChanged = QtCore.pyqtSignal(str,str)
    @staticmethod
    def naIfEmpty(value):
        if not value:
            return "N/A"
        return value

    def __init__(self, parent=None):
        super(SelectCanPortDialog, self).__init__(parent)
        self.setWindowModality(QtCore.Qt.ApplicationModal)
        self.setWindowTitle("Select a Port")
        
        self.resize(500, 400)

        self.portsList = QtWidgets.QTreeWidget()
        openButton = QtWidgets.QPushButton("Open")
        openButton.clicked.connect(self.openPort)

        createButton = QtWidgets.QPushButton("Create")
        createButton.clicked.connect(self.createPort)

        deleteButton = QtWidgets.QPushButton("Delete")
        deleteButton.clicked.connect(self.deletePort)
        
        self._arbBitrateEdit = QtWidgets.QLineEdit("1,000,000")
        self._dataBitrateEdit = QtWidgets.QLineEdit("5,000,000")
        configureButton = QtWidgets.QPushButton("Configure")
        configureButton.clicked.connect(self.configurePort)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.portsList)
        layout.addWidget(createButton)
        layout.addWidget(openButton)
        layout.addWidget(deleteButton)
        
        hLayout = QtWidgets.QHBoxLayout()
        hLayout.addWidget(QtWidgets.QLabel("Arbitration Bitrate"))
        hLayout.addWidget(self._arbBitrateEdit)
        hLayout.addWidget(QtWidgets.QLabel("Data Bitrate"))
        hLayout.addWidget(self._dataBitrateEdit)
        hLayout.addWidget(configureButton)
        layout.addLayout(hLayout)

        self.setLayout(layout)
        
        self._socketcan = SocketcanCommands(self)
        self._socketcan.statusUpdate.connect(self.statusUpdate)
        self._socketcan.portsChanged.connect(self.setup_list)
        
        self.setup_list()
    
    # This is annoying, but we have to tell the SocketcanCommands object to stop,
    # otherwise it gets deleted with it's QProcess still running.
    def closeEvent(self, event):
        self._socketcan.stop()

    def setup_list(self):
        self.portsList.clear()
        
        buses = self._socketcan.list()
        
        if len(buses) > 0:
            tableHeader = list(buses[0].keys())
            self.portsList.setHeaderLabels(tableHeader)
            for info in buses:
                tableElems = [str(x) for x in info.values()]
                self.portsList.addTopLevelItem(QtWidgets.QTreeWidgetItem(None, tableElems))
        else:
            tableHeader = ["interface", "channel"]
            self.portsList.setHeaderLabels(tableHeader)
        create_item = QtWidgets.QTreeWidgetItem(None, ["new vcan", "[NAME]"])
        create_item.setFlags(create_item.flags() | QtCore.Qt.ItemIsEditable)

        self.portsList.addTopLevelItem(create_item)

        for i in range(0, len(tableHeader)):
            self.portsList.resizeColumnToContents(i)
    
    def createPort(self):
        cur_item = self.portsList.currentItem()
        self._socketcan.create(cur_item.text(1))

    def deletePort(self):
        cur_item = self.portsList.currentItem()
        self._socketcan.delete(cur_item.text(1))

    def configurePort(self):
        cur_item = self.portsList.currentItem()
        arb_bitrate = int(self._arbBitrateEdit.text().replace(",",""))
        data_bitrate = int(self._dataBitrateEdit.text().replace(",",""))
        self._socketcan.configure(cur_item.text(1), arb_bitrate, data_bitrate)

    def openPort(self):
        cur_item = self.portsList.currentItem()
        if cur_item is not None:
            interface = cur_item.text(0)
            channel = cur_item.text(1)
            self.portChanged.emit(interface, channel)
            self.close()




#!/usr/bin/env python3
import os
import sys
import argparse
from PyQt5 import QtGui, QtWidgets, QtCore
import time

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)
from msgtools.lib.messaging import Messaging
import msgtools.lib.gui

DESCRIPTION='''
    BandwidthTester is designed to work with the BandWidthTestEcho utility.  It connects to a
    MsgServer and sends Bandwidth test messages, listening for echo responses with timestamps.
    Basic stats are provided to allow you rough order of throughput.  Python is not very stable
    when it comes to timing and multi-threaded operations so expect some margin of error in your 
    results.
'''

class BandwidthTester(msgtools.lib.gui.Gui):
    startTime = int(time.time() * 1000)
    def __init__(self, parent=None):
        parser = argparse.ArgumentParser(description=DESCRIPTION)
        parser = msgtools.lib.gui.Gui.addBaseArguments(parser)
        args = parser.parse_args()

        msgtools.lib.gui.Gui.__init__(self, "Bandwidth Tester 0.1", args, parent)
        
        self.lastTxSequence = 0
        self.lastRxSequence = 0
        self.txMsgCount = 0
        self.txLen = 0
        self.txByteCount = 0
        self.rxByteCount = 0
        self.rxBytesPerSec = 0
        self.txBytesPerSec = 0
        self.totalLatency = 0
        self.latencySamples = 0
        
        for msgClassName in Messaging.MsgClassFromName:
            if "BandwidthTest" in msgClassName:
                self.bandwidthTestMsgClass = Messaging.MsgClassFromName[msgClassName]
                break
        self.maxSeq = int(self.bandwidthTestMsgClass.GetSequenceNumber.maxVal)
        
        # event-based way of getting messages
        self.RxMsg.connect(self.ProcessMessage)

        self.msgTimer = QtCore.QTimer(self)
        self.msgTimer.setInterval(10)
        self.msgTimer.timeout.connect(self.msgTimeout)

        self.displayTimer = QtCore.QTimer(self)
        self.displayTimer.setInterval(1000)
        self.displayTimer.timeout.connect(self.updateDisplay)
        self.displayTimer.start()

        vbox = QtWidgets.QVBoxLayout()
        centralWidget = QtWidgets.QWidget()
        centralWidget.setLayout(vbox)
        self.setCentralWidget(centralWidget)
        
        hbox = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox)
        hbox.addWidget(QtWidgets.QLabel("Tx bytes"))
        self.txByteCountLabel = QtWidgets.QLabel()
        hbox.addWidget(self.txByteCountLabel)
        hbox.addWidget(QtWidgets.QLabel("Rx bytes"))
        self.rxByteCountLabel = QtWidgets.QLabel()
        hbox.addWidget(self.rxByteCountLabel)

        hbox = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox)
        hbox.addWidget(QtWidgets.QLabel("Tx bytes/sec"))
        self.txBytesPerSecLabel = QtWidgets.QLabel()
        hbox.addWidget(self.txBytesPerSecLabel)
        hbox.addWidget(QtWidgets.QLabel("Rx bytes/sec"))
        self.rxBytesPerSecLabel = QtWidgets.QLabel()
        hbox.addWidget(self.rxBytesPerSecLabel)
        
        hbox = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox)
        self.latencyLabel = QtWidgets.QLabel()
        hbox.addWidget(QtWidgets.QLabel("Round-trip latency"))
        hbox.addWidget(self.latencyLabel)
        hbox.addWidget(QtWidgets.QLabel("ms"))

        clearBtn = QtWidgets.QPushButton(self)
        clearBtn.setText('Clear')
        clearBtn.clicked.connect(self.clearStats)
        vbox.addWidget(clearBtn)
        
        timerSlider = QtWidgets.QSlider(QtCore.Qt.Horizontal, self)
        timerSlider.valueChanged.connect(self.msgTimer.setInterval)
        timerLabel = QtWidgets.QLabel()
        timerSlider.valueChanged.connect(lambda newVal: timerLabel.setText(str(newVal)+" ms"))
        timerSlider.setMinimum(10)
        timerSlider.setMaximum(1000)
        timerSlider.setValue(500)
        hbox = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox)
        hbox.addWidget(QtWidgets.QLabel("Timer (ms)"))
        hbox.addWidget(timerSlider)
        hbox.addWidget(timerLabel)
        
        txMsgCountSlider = QtWidgets.QSlider(QtCore.Qt.Horizontal, self)
        txMsgCountSlider.valueChanged.connect(self.setTxMsgCount)
        msgCountLabel = QtWidgets.QLabel()
        txMsgCountSlider.valueChanged.connect(lambda newVal: msgCountLabel.setText(str(newVal)+" msgs"))
        txMsgCountSlider.setMinimum(1)
        txMsgCountSlider.setMaximum(100)
        hbox = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox)
        hbox.addWidget(QtWidgets.QLabel("Msg Count"))
        hbox.addWidget(txMsgCountSlider)
        hbox.addWidget(msgCountLabel)

        txDataLenSlider = QtWidgets.QSlider(QtCore.Qt.Horizontal, self)
        txDataLenSlider.valueChanged.connect(self.setTxLen)
        txDataLenLabel = QtWidgets.QLabel()
        txDataLenSlider.valueChanged.connect(lambda newVal: txDataLenLabel.setText(str(newVal*int(self.bandwidthTestMsgClass.GetTestData.size)+int(self.bandwidthTestMsgClass.GetTestData.offset))+" bytes"))
        txDataLenSlider.setMinimum(0)
        txDataLenSlider.setMaximum(int(self.bandwidthTestMsgClass.GetTestData.count))
        txDataLenSlider.setValue(3)
        hbox = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox)
        hbox.addWidget(QtWidgets.QLabel("Msg Len"))
        hbox.addWidget(txDataLenSlider)
        hbox.addWidget(txDataLenLabel)

        self.startStop = QtWidgets.QPushButton(self)
        self.startStop.setText('Stop')
        self.startStop.clicked.connect(self.startStopFn)
        vbox.addWidget(self.startStop)
        
        self.msgTimer.start()
        
    def startStopFn(self):
        if self.msgTimer.isActive():
            self.msgTimer.stop()
            self.startStop.setText('Start')
        else:
            self.msgTimer.start()
            self.startStop.setText('Stop')
    
    def setTxMsgCount(self, txCount):
        self.txMsgCount = txCount

    def setTxLen(self, txLen):
        self.txLen = txLen

    def msgTimeout(self):
        bytesTransmitted = 0
        for i in range(0, self.txMsgCount):
            msg = self.bandwidthTestMsgClass()
            self.lastTxSequence += 1
            if self.lastTxSequence > self.maxSeq:
                self.lastTxSequence = 0
            msg.SetSequenceNumber(self.lastTxSequence)
            msgLen = int(msg.SetTestData.offset) + int(msg.SetTestData.size)*self.txLen
            if(self.txLen >= 1):
                sendTime = int(time.time() * 1000)-self.startTime
                msg.SetTestData(sendTime,0)
            msg.hdr.SetDataLength(msgLen)
            self.txByteCount += msgLen
            self.txBytesPerSec += msgLen
            self.SendMsg(msg)
        self.txByteCountLabel.setText(str(self.txByteCount))

    def ProcessMessage(self, msg):
        if type(msg) == self.bandwidthTestMsgClass:
            desiredSeq = self.lastRxSequence+1
            if desiredSeq > self.maxSeq:
                desiredSeq = 0
            if msg.GetSequenceNumber() == desiredSeq:
                if msg.hdr.GetDataLength() >= int(msg.GetTestData.offset)+int(msg.GetTestData.size):
                    sendTime = msg.GetTestData(0)
                    recvTime = int(time.time() * 1000)-self.startTime
                    latency = recvTime - sendTime
                    self.totalLatency += latency
                    self.latencySamples += 1
                self.rxByteCount += msg.hdr.GetDataLength()
                self.rxBytesPerSec += msg.hdr.GetDataLength()
            else:
                print("ERROR!  Got sequence " + str(msg.GetSequenceNumber()) + ", but wanted " + str(self.lastRxSequence))
            self.lastRxSequence = msg.GetSequenceNumber()
            self.rxByteCountLabel.setText(str(self.rxByteCount))
    
    def clearStats(self):
        self.txByteCount = 0
        self.rxByteCount = 0
        self.totalLatency = 0
        self.latencySamples = 0
        self.txByteCountLabel.setText("0")
        self.rxByteCountLabel.setText("0")
        self.latencyLabel.setText("")

    
    def updateDisplay(self):
        self.rxBytesPerSecLabel.setText(str(self.rxBytesPerSec))
        self.txBytesPerSecLabel.setText(str(self.txBytesPerSec))
        if self.latencySamples > 0:
            self.latencyLabel.setText('%.2f' % (self.totalLatency / self.latencySamples))
        else:
            self.latencyLabel.setText("undef")
        self.rxBytesPerSec = 0
        self.txBytesPerSec = 0
        self.totalLatency = 0
        self.latencySamples = 0

def main(args=None):
    app = QtWidgets.QApplication(sys.argv)
    msgApp = BandwidthTester()
    msgApp.show()
    sys.exit(app.exec_())

# main starts here
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import json
import pandas as pd
import time
from collections import OrderedDict

from msgtools.lib.file_reader import MessageFileReader
from msgtools.lib.messaging import Messaging

# function to flatten a dictionary of a Message object, according to
# some unique properties of Message objects.
def flatten_msg(msg):
    out = {}

    def flatten(x, name=''):
        if type(x) is dict or type(x) is OrderedDict:
            for a in x:
                # leave the header as a dictionary, otherwise it's a lot of clutter
                if name == '' and a == "hdr":
                    #print(name+a)
                    # Put the whole header in the output
                    #out[name+a] = x[a]

                    # make other hdr fields a regular field
                    for field in x[a].keys():
                        if "time" in field.lower():
                            out[name+field] = x[a][field]
                        elif field in ["ID", "DataLength"]:
                            pass
                        elif field in ["Source", "Destination"]:
                            if x[a][field] != 0:
                                out[name+field] = x[a][field]
                else:
                    flatten(x[a], name + a + '.')
        elif type(x) is list:
            i = 0
            for a in x:
                flatten(a, name[:-1] + "["+str(i) + '].') # extra . will be removed by next call to flatten()
                i += 1
        else:
            out[name[:-1]] = x

    flatten(msg)
    return out

def load_json(filename):
    dict_of_dataframes = {}

    with open(filename) as f:
        for line in f:
            j = json.loads(line)
            for msgname in j.keys():
                # "flat" is a dictionary, but not nested!
                flat = flatten_msg(j[msgname])

                # add to dictionary of dataframes
                if not msgname in dict_of_dataframes:
                    dict_of_dataframes[msgname] = []
                dict_of_dataframes[msgname].append(flat)

    for msgname in dict_of_dataframes:
        if "FIELD_METADATA." in msgname:
            # message metadata, occurs only once for each message type, with no time tag.
            dict_of_dataframes[msgname] = pd.DataFrame.from_records(dict_of_dataframes[msgname], index="name")
        else:
            # normal messages, with time-tagged data
            dict_of_dataframes[msgname] = pd.DataFrame.from_records(dict_of_dataframes[msgname], index="Time")

    return dict_of_dataframes

class PandasBinaryReader(MessageFileReader):
    def __init__(self, filename, serial):
        super(PandasBinaryReader, self).__init__()
        self.dict_of_dataframes = {}
        header_name = None
        if serial:
            header_name = "SerialHeader"
        self.read_file(filename, header_name)

        for msgname in self.dict_of_dataframes:
            self.dict_of_dataframes[msgname] = pd.DataFrame.from_records(self.dict_of_dataframes[msgname], index="Time")
        

    def process_message(self, msg):
        # Get a dictionary of the message
        d = msg.toDict(includeHeader=True)

        # the dictionary has one key, which is the name of the message
        msgname = list(d.keys())[0]

        flat = flatten_msg(d[msgname])

        # add to dictionary of dataframes
        if not msgname in self.dict_of_dataframes:
            self.dict_of_dataframes[msgname] = []
        self.dict_of_dataframes[msgname].append(flat)

def load_binary(filename, serial=None):
    # if user didn't specify value for serial, set it based on filename.
    if serial == None:
        if filename.endswith(".bin") or filename.endswith(".log"):
            serial = False
        elif filename.endswith(".txt"):
            serial = True

    # Need to read binary messages and process them.
    # flatten_msg() above was written to operate on python dicts returned
    # by reading JSON.  The Message class already overloads the __getattr__()
    # function to make field access work like attribute access, and perhaps
    # if we also override __dir__(), then Message objects could be operated
    # on like python objects with named attributes.  If that doesn't work,
    # we can either change flatten_msg() to work on Message objects, using their
    # normal means of reflection, or generate JSON and then parse the JSON to
    # form a dictionary, and then call flatten_msg() on that dictionary.
    pandas_reader = PandasBinaryReader(filename, serial)

    return pandas_reader.dict_of_dataframes

def field_units(dict_of_dataframes, msgname, fieldname):
    return dict_of_dataframes["FIELD_METADATA."+msgname].loc[fieldname]["units"]

# Make a subclass of pandas.DataFrame that doesn't raise exceptions when
# plot is called, even if there's no data to plot.
class DataFrameSubclass(pd.DataFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.field_metadata = None
    def plot(self, *args, **kwargs):
        # Build a mapping for how to rename columns to include units
        rename_with_units = {}
        for c in self.columns.values:
            units = self.field_metadata.loc[c]["units"]
            if units != "":
                rename_with_units[c] = "%s (%s)" % (c, units)

        # Plot the columns, with the new column names that include units.
        try:
            super().rename(columns=rename_with_units).plot(*args, **kwargs)
        except TypeError:
            # If there's nothing to plot, print a message instead of crashing.
            print("Not plotting %s%s, nothing to plot!" % (self.msgname, self.columns.values))

# Make a subclass of dict that has a function that returns an empty dataframe
# if there's no matching dataframe with the specified keys.
class DataFrameDict(dict):
    def data(self, msgname, fieldnames):
        try:
            msg_dataframe = self[msgname]
        except KeyError:
            print("ERROR! [%s][%s] not in dataframe!" % (msgname, fieldnames))
            ret = DataFrameSubclass(columns=fieldnames)
            ret.msgname = msgname
            return ret
        
        # if we didn't hit the exception to return an empty DataFrameSubclass,
        # return a DataFrameSubclass of the dataframe we care about, but add
        # some extra members to it to access field metadata from the dataframe
        # dictionary.
        ret = DataFrameSubclass(msg_dataframe[fieldnames])
        ret.field_metadata = self["FIELD_METADATA."+msgname]
        return ret

    def field_units(self, msgname, fieldname):
        return self["FIELD_METADATA."+msgname].loc[fieldname]["units"]

# A global variable for the last filename that was loaded.
# This is useful for when None is passed in to load(),
# and the user wants to know what file was auto-loaded.
last_filename = None

# Return a hashtable of pandas dataframes from the file that was loaded.
def load(filename=None, serial=None):
    # if the filename was an int, use it as an index into the list of files,
    # and reset filename to None.  If filename was a valid file name, then
    # the cast to int will raise an exception and filename won't be reset.
    fileindex = 0
    try:
        fileindex = int(filename)
        filename = None
    except:
        pass
    if filename == None:
        import os, re
        def filename_is_log(filename):
            if not os.path.isfile(filename):
                return False
            # the filename should contain something that looks like YYYYMMDD in it
            if not re.search(pattern=r'20\d{6}[-_]\d{6}', string=filename):
                return False
            filename_elements = os.path.splitext(filename)
            if len(filename_elements) < 2:
                return False
            ext = filename_elements[1]
            if not ext in [".json", ".bin", ".log"]:
                return False
            return True
        def log_files(dir):
            return [dir+"/"+f for f in os.listdir(dir) if filename_is_log(dir+"/"+f)]
        def timestring(filename):
            return re.search(pattern=r'20\d{6}[-_]\d{6}', string=filename).group()
        # Get log files in current dir
        filenames = log_files(".")
        # add log files in logs subdir, if it exists
        try:
            filenames += log_files("logs")
        except FileNotFoundError:
            pass
        # sort by the timestring that was found in the filename
        filenames.sort(reverse=True, key=lambda x: timestring(x))
        filename = filenames[fileindex]

    global last_filename
    last_filename = filename

    if filename.endswith(".json"):
        ret = load_json(filename)
    elif filename.endswith(".bin") or filename.endswith(".log") or filename.endswith(".txt"):
        ret = load_binary(filename, serial)
    
    return DataFrameDict(ret)

# Make this script executable from the shell, with command line arguments.
# This is just for testing, since there's not much point in printing a DataFrame
# to stdout.
def main():
    import argparse
    parser = argparse.ArgumentParser(description="Read a log file into pandas DataFrames")
    parser.add_argument('filename', nargs="?", default=None, help='''The log file you want to split into CSV.  
        .log extension assumes the log is binary with NetworkHeaders.  A .txt extension assumes the 
        file was created with SerialHeaders.''')
    parser.add_argument('--serial', action='store_true', help='''Assumes input file contains binary messages with SerialHeaders instead of NetworkHeaders.''')
    args = parser.parse_args()

    if args.filename and args.filename.lower().endswith('.txt'):
        args.serial = True
    
    df = load(args.filename, args.serial)
    print(df)

# main starts here
if __name__ == '__main__':
    main()



from .messaging import Messaging

# for conversion to JSON
from collections import OrderedDict
import json
import struct

def toDict(msg, includeHeader=False):
    pythonObj = OrderedDict()
    if includeHeader:
        pythonObj['hdr'] = OrderedDict()
        for fieldInfo in msg.hdr.fields:
            if(fieldInfo.count == 1):
                if len(fieldInfo.bitfieldInfo) == 0:
                    pythonObj['hdr'][fieldInfo.name] = Messaging.get(msg.hdr, fieldInfo)
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        pythonObj['hdr'][bitInfo.name] = Messaging.get(msg.hdr, bitInfo)
            else:
                arrayList = []
                terminate = 0
                for i in range(0,fieldInfo.count):
                    arrayList.append(Messaging.get(msg.hdr, fieldInfo, i))
                pythonObj['hdr'][fieldInfo.name] = arrayList

    msgClass = Messaging.MsgClass(msg.hdr)
    for fieldInfo in msgClass.fields:
        if(fieldInfo.count == 1):
            if not fieldInfo.exists(msg):
                break
            if len(fieldInfo.bitfieldInfo) == 0:
                pythonObj[fieldInfo.name] = Messaging.get(msg, fieldInfo)
            else:
                for bitInfo in fieldInfo.bitfieldInfo:
                    pythonObj[bitInfo.name] = Messaging.get(msg, bitInfo)
        else:
            if len(fieldInfo.bitfieldInfo) == 0:
                arrayList = []
                terminate = 0
                for i in range(0,fieldInfo.count):
                    if not fieldInfo.exists(msg, i):
                        terminate = 1
                        break
                    arrayList.append(Messaging.get(msg, fieldInfo, i))
                pythonObj[fieldInfo.name] = arrayList
                if terminate:
                    break
            else:
                # Arrays-of-structs acting like parallel arrays.
                for bitInfo in fieldInfo.bitfieldInfo:
                    arrayList = []
                    for i in range(0,fieldInfo.count):
                        if not fieldInfo.exists(msg, i):
                            break
                        arrayList.append(Messaging.get(msg, bitInfo, i))
                    pythonObj[bitInfo.name] = arrayList

    return {msg.MsgName() : pythonObj}

def toJson(msg, includeHeader=False):
    return json.dumps(toDict(msg,includeHeader))

def jsonToMsg(jsonString, ignore_invalid=False):
    d = json.loads(jsonString)
    return dictToMsg(d, ignore_invalid)

def dictToMsg(d, ignore_invalid=False):
    terminationLen = None
    def handle_header(hdrDict):
        length = None
        for fieldName in hdrDict:
            if fieldName == "DataLength":
                if hdrDict[fieldName] == ";":
                    length = 0
                else:
                    length = int(hdrDict[fieldName])
        return length
    if "hdr" in d:
        hdrDict = d["hdr"]
        terminationLen = handle_header(hdrDict)
    for msgName in d:
        if msgName == "hdr":
            # hdr handled above, *before* message body
            pass
        else:
            fieldDict = d[msgName]
            msgClass = Messaging.MsgClassFromName[msgName]
            msg = msgClass()
            for fieldName in fieldDict:
                if fieldName == "hdr":
                    if "Time" in fieldDict["hdr"]:
                        msg.hdr.SetTime(fieldDict["hdr"]["Time"])
                    terminationLen = handle_header(fieldDict["hdr"])
                    continue
                fieldInfo = Messaging.findFieldInfo(msgClass.fields, fieldName)
                if fieldInfo == None:
                    if ignore_invalid:
                        if ignore_invalid != "silent":
                            print("Ignoring invalid field %s.%s" % (msgName, fieldName))
                        continue
                    else:
                        raise KeyError("Invalid field %s for message %s" % (fieldName, msgName))
                fieldValue = fieldDict[fieldName]
                if isinstance(fieldValue, list):
                    #print(fieldName + " list type is " + type(fieldValue))
                    for i in range(0,len(fieldValue)):
                        try:
                            Messaging.set(msg, fieldInfo, fieldValue[i], i)
                        except struct.error as e:
                            print(e)
                            break
                        if terminationLen != None:
                            #TODO Broken for arrays-of-structs acting like parallel arrays!
                            terminationLen = max(terminationLen, fieldInfo.end_location(i))
                elif isinstance(fieldValue, dict):
                    #print(fieldName + " dict type is " + str(type(fieldValue)))
                    if fieldInfo.bitfieldInfo:
                        pass
                    else:
                        pass
                else:
                    #print(str(type(fieldValue)) + " " + fieldName + ", calling set with " + str(fieldValue))
                    Messaging.set(msg, fieldInfo, fieldValue)
                    if terminationLen != None:
                        if fieldInfo.type == "string":
                            #TODO Broken for arrays-of-structs acting like parallel arrays!
                            terminationLen = max(terminationLen, fieldInfo.end_location(len(fieldValue)-1))
                        else:
                            #TODO Broken for arrays-of-structs acting like parallel arrays!
                            terminationLen = max(terminationLen, fieldInfo.end_location())
    # Shorten the message if necessary.
    if terminationLen != None and terminationLen < msg.hdr.GetDataLength():
        msg.hdr.SetDataLength(terminationLen)
    return msg

# return list of metadata for fields
def jsonHeader(msg):
    ret = ""
    msgClass = Messaging.MsgClass(msg.hdr)
    for fieldInfo in msgClass.fields:
        e = None
        if len(fieldInfo.enum) > 0:
            e = fieldInfo.enum[0] # list element 0 is forward mapping (list element 1 is reverse mapping)
        pythonObj = OrderedDict()
        pythonObj["name"]=fieldInfo.name
        pythonObj["type"]=fieldInfo.type
        pythonObj["units"]=fieldInfo.units
        pythonObj["min"]=fieldInfo.minVal
        pythonObj["max"]=fieldInfo.maxVal
        pythonObj["description"]=fieldInfo.description
        pythonObj["count"]=fieldInfo.count
        #pythonObj["enum"]=e
        ret += json.dumps({"FIELD_METADATA."+msg.MsgName() : pythonObj}) + "\n"
        if len(fieldInfo.bitfieldInfo) > 0:
            for bitInfo in fieldInfo.bitfieldInfo:
                e = None
                if len(bitInfo.enum) > 0:
                    e = bitInfo.enum[0] # list element 0 is forward mapping (list element 1 is reverse mapping)
                pythonObj = OrderedDict()
                pythonObj["name"]=bitInfo.name
                pythonObj["type"]=bitInfo.type
                pythonObj["units"]=bitInfo.units
                pythonObj["min"]=bitInfo.minVal
                pythonObj["max"]=bitInfo.maxVal
                pythonObj["description"]=bitInfo.description
                pythonObj["count"]=bitInfo.count
                #pythonObj["enum"]=e

                ret += json.dumps({"FIELD_METADATA."+msg.MsgName() : pythonObj}) + "\n"
    return ret


import copy
import gevent
from gevent.event import Event
import logging
import time

# This is used for formatting logging messages.
class StyleAdapter(logging.LoggerAdapter):
    def __init__(self, logger, sim):
        self.logger = logger
        self.sim = sim

    def log(self, level, msg, *args):
        if self.isEnabledFor(level):
            t = SimExec.time()
            self.logger._log(level, ("%7.3f: " % t) + msg, args, ())

# This class is used to execute a simulation, and allow tasks to be run along
# with it, at the appropriate simulation time rate.
# If no simulation exists, it falls back to using Python system time instead
# of simulation time.
class SimExec:
    # a gevent.event used to get coroutines to run when simulation time increases
    tick_event = None
    # a boolean to indicate if a simulation really exists.
    # if not, users probably want to use Python time instead of simulation time
    sim_exists = False
    allow_socket_comms = True
    _current_time = 0.0
    logging = StyleAdapter(logging.getLogger(__name__), None)
    def __init__(self, dt, args, logging):
        SimExec.tick_event = Event()
        self._args = args
        # If running ASAP, don't allow socket communications, because
        # it'd slow things down too much, and if any socket clients
        # were connected, they wouldn't be able to handle the flood of
        # messages.  ASAP is only good for batch processing with data
        # being written to files.
        if self._args.asap:
            SimExec.allow_socket_comms = False
        self.logging = logging
        SimExec.logging = logging
        self.dt = dt
        self.iterations_per_second = int(1/self.dt)
        self.start_time = time.time()
        self.last_time = self.start_time
        self.iterations = 0
        self.incremental_sim_time = 0
        self.incremental_wallclock_time = 0
        self.incremental_execution_time = 0

        # for lockstep mode
        self.ready_to_run = True

    @staticmethod
    def time():
        # Use SimExec.sim_exists to tell if a sim object exists, so we know if time
        # needs to sync to sim time, or if it should be free-running system time.
        if SimExec.sim_exists:
            return SimExec._current_time
        else:
            return time.time()

    # wait for some amount of simulation time (or real time, if sim not running)
    @staticmethod
    def wait(dt):
        # Use SimExec.sim_exists to tell if a sim object exists, so we know if time
        # needs to sync to sim time, or if it should be free-running system time.
        if SimExec.sim_exists:
            t1 = SimExec._current_time
            if dt > 0.0:
                while True:
                    SimExec.tick_event.wait()
                    SimExec.tick_event.clear()
                    t2 = SimExec._current_time
                    if t2 - t1 > dt:
                        break
                    #print("not yet")
            else:
                # sleep for zero time, to allow another coroutine to run
                gevent.sleep(0)
                t2 = SimExec._current_time
        else:
            t1 = time.time()
            #time.sleep(dt)
            # The Client implementation changed to always use greenlets, so we
            # should *always* do a gevent.sleep() here!
            # Undo this if the Client changes so that it doens't use greenlets
            # when SimExec.sim_exists is False!
            gevent.sleep(dt)
            t2 = time.time()
        SimExec.logging.debug("SimExec.wait(%.3f ms): %.3f ms" % (1000.0*dt, 1000.0*(t2-t1)))
        #print(f"{i} waited {dt} seconds")

    def time_tick(self, simtime):
        SimExec._current_time = simtime
        # perform a gevent Event set, to wake anything waiting on sim
        # time changing.
        SimExec.tick_event.set()
        now = time.time()
        actual_dt = now - self.last_time
        self.iterations += 1
        self.incremental_sim_time += self.dt
        if self._args.asap:
            self.incremental_wallclock_time += actual_dt
            if self.iterations % self.iterations_per_second == 0:
                self.logging.info("simtime %.1f / walltime %.3f = %.1fx acceleration" % (self.incremental_sim_time, self.incremental_wallclock_time, self.incremental_sim_time/self.incremental_wallclock_time))
                self.incremental_sim_time = 0.0
                self.incremental_wallclock_time = 0.0
            # do zero-time sleep, so other coroutines can run if they had yielded
            gevent.sleep(0)
            self.last_time = now
        else:
            self.incremental_execution_time += (now-self.last_time)
            sleep_time = self.dt - actual_dt
            self.logging.log(logging.ERROR if sleep_time < 0.25 * self.dt else logging.DEBUG, "Execution took %f seconds, want to run every %f seconds, sleeping %f seconds" % (actual_dt, self.dt, sleep_time))
            if self.iterations % self.iterations_per_second == 0:
                self.logging.warning("cputime %.3f / walltime %.1f = %.2f%% processor usage" % (self.incremental_execution_time, self.incremental_sim_time, 100.0*(self.incremental_execution_time/self.incremental_sim_time)))
                self.incremental_sim_time = 0.0
                self.incremental_wallclock_time = 0.0
                self.incremental_execution_time = 0.0
            if self.dt > actual_dt:
                gevent.sleep(sleep_time)
            else:
                # do zero-time sleep, so other coroutines can run if they had yielded
                gevent.sleep(0)
            self.last_time = time.time()



import ctypes
from msgtools.lib.messaging import *
import msgtools.lib.msgjson as msgjson
import msgtools.lib.msgcsv as msgcsv
import collections

# This class is used so that individual elements of array fields can be
# accessed using array index notation [] on the field of the message.
# __getattr__ on the Message object needs to return a FieldListAccessor
# which can get/set individual elements of the message.
class FieldListAccessor(collections.UserList):
    def __init__(self, msg, getfn, setfn):
        self._msg = msg
        self._setfn = setfn
        self._getfn = getfn

    def __repr__(self):
        ret = []
        for i in range(0, self._setfn.count):
            try:
                ret.append(self._getfn(i))
            except struct.error:
                pass
        return str(ret)

    def __getitem__(self, key):
        return self._getfn(key)

    def __setitem__(self, key, value):
        self._setfn(value, key)

class Message:
    def __init__(self, messageBuffer=None, id=None, size=None):
        doInit = 0
        if messageBuffer == None:
            doInit = 1
            messageBuffer = ctypes.create_string_buffer(Messaging.hdrSize + size)
        else:
            try:
                messageBuffer.raw
            except AttributeError:
                newbuf = ctypes.create_string_buffer(len(messageBuffer))
                for i in range(0, len(messageBuffer)):
                    newbuf[i] = bytes(messageBuffer)[i]
                messageBuffer = newbuf
        # this is a trick to get us to store a copy of a pointer to a buffer, rather than making a copy of the buffer
        self.msg_buffer_wrapper = { "msg_buffer": messageBuffer }
        
        # this is a trick to allow the creation of fake fields that aren't actually in a message,
        # but that we want to act like fields of the message
        self.fake_fields = {}

        self.hdr = Messaging.hdr(messageBuffer)

        if doInit:
            self.hdr.SetMessageID(id)
            self.hdr.SetDataLength(size)

    def rawBuffer(self):
        # this is a trick to get us to store a copy of a pointer to a buffer, rather than making a copy of the buffer
        return self.msg_buffer_wrapper["msg_buffer"]

    def set_fields(self, **kwargs):
        for param, value in kwargs.items():
            fn = getattr(self, "Set"+param)
            if fn.count > 1:
                if not isinstance(value, list):
                    value = [value]
                for idx in range(0,len(value)):
                    fn(value[idx], idx)
            else:
                fn(value)

    def __getattr__(self, attr):
        fn_name = 'Get' + attr
        if hasattr(type(self), fn_name):
            fn = getattr(self,fn_name)
            if callable(fn):
                if fn.count > 1:
                    setfn = getattr(self,"Set"+attr)
                    return FieldListAccessor(self, fn, setfn)
                else:
                    return fn()
        elif attr in self.fake_fields:
            return self.fake_fields[attr]
        raise AttributeError('Message %s has no field %s' % (self.MsgName(), attr))

    def __setattr__(self, attr, val):
        fn_name = 'Set' + attr
        if (not attr.startswith('Set')) and (not attr.startswith('Get')) and fn_name in dir(type(self)):
            fn = getattr(self,fn_name)
            if callable(fn):
                if fn.count > 1:
                    if not isinstance(val, list):
                        raise AttributeError('Message %s field %s needs array param, not %s' % (self.MsgName(), attr, str(val)))
                    for i in range(0,len(val)):
                        try:
                            fn(val[i], i)
                        except struct.error:
                            break
                    return
                else:
                    fn(val)
                    return
        allowed_attributes = ['msg_buffer_wrapper', 'hdr', 'fake_fields']
        if attr in allowed_attributes:
            super(Message, self).__setattr__(attr, val)
        else:
            raise AttributeError('Message %s has no field %s' % (self.MsgName(), attr))

    @staticmethod
    def fromJson(s, ignore_invalid=False):
        return msgjson.jsonToMsg(s, ignore_invalid)

    @staticmethod
    def fromCsv(s):
        return msgcsv.csvToMsg(s)
    
    # we should add support for msgname(p1,p2,p3, [p4a, p4b, p4c])
    # and also msgname(f1=p1, f3=p3, f4=[p4a, p4b, p4c], f2=p2)
    # it'll be useful because it's the exact same syntax as the message constructor takes,
    # so will make it possible for the msgtools.lib.gui.LineEditWithHistory to use same
    # syntax as python scripts do.  it also matches what Message.__str__ outputs.
    
    # alternatively, should we have one fromStr function that accepts all of:
    # 1) json
    # 2) csv
    # 3) string that works as python object constructor
    # ?
    @staticmethod
    def fromStr(s):
        pass

    def toDict(self, includeHeader=False):
        return msgjson.toDict(self, includeHeader)

    def toJson(self, includeHeader=False):
        return msgjson.toJson(self, includeHeader)

    def toCsv(self, nameColumn=True, timeColumn=False):
        return msgcsv.toCsv(self, nameColumn=nameColumn, timeColumn=timeColumn)
    
    def csvHeader(self, nameColumn=True, timeColumn=False):
        return msgcsv.csvHeader(self, nameColumn=nameColumn, timeColumn=timeColumn)

    def jsonHeader(self):
        return msgjson.jsonHeader(self)

    def __repr__(self):
        def add_param(n, v):
            if v == '':
                v = '""'
            if ',' in v and not((v.startswith('"') and v.endswith('"')) or (v.startswith("'") and v.endswith("'"))):
                v = '"%s"' % v
            return n + " = " + v + ", "
        ret = ''
        for fieldInfo in self.fields:
            if(fieldInfo.count == 1):
                if not fieldInfo.exists(self):
                    break
                if len(fieldInfo.bitfieldInfo) == 0:
                    ret += add_param(fieldInfo.name, str(Messaging.get(self, fieldInfo)))
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        ret += add_param(bitInfo.name, str(Messaging.get(self, bitInfo)))
            else:
                arrayList = []
                terminate = 0
                for i in range(0,fieldInfo.count):
                    if not fieldInfo.exists(self, i):
                        terminate = 1
                        break
                    arrayList.append(str(Messaging.get(self, fieldInfo, i)))
                ret += add_param(fieldInfo.name, "[" + ','.join(arrayList) + "]")
                if terminate:
                    break

        if ret.endswith(", "):
            ret = ret[:-2]
        return "%s(%s)" % (self.MsgName(), ret)



#!/usr/bin/env python3
import collections
import hashlib
import json
import os
import re
import sys
import string
import shutil

def Usage():
    sys.stderr.write('Usage: ' + sys.argv[0] + ' inputdir dictionaryFile dictionaryHeaderFile\n')
    sys.exit(1)

def ProcessDir(inputDir):
    for filename in sorted(os.listdir(inputDir)):
        inputFilename = inputDir + '/' + filename
        if os.path.isdir(inputFilename):
            ProcessDir(inputFilename)
        else:
            if filename.endswith(".h") or filename.endswith(".c") or filename.endswith(".cpp"):
                if filename != "debug_printf.h" and filename != "debug_printf.cpp":
                    #print("reading " + inputFilename)
                    ProcessFile(inputFilename.replace("//", "/"))

def theSameWithoutComments(file1, file2):
    with open(file1) as text1, open(file2) as text2:
        if [line for line in text1 if not line.startswith('#')] == [line for line in text2 if not line.startswith('#')]:
            return True
    return False
    
def printDictionary(dictFilename, headerFilename, dictionaryDeployDir):
    # Read in the old dictionary file so we can compare it to new desired
    # contents, to decide if we need to rewrite it.
    try:
        with open(dictFilename, 'r') as file:
            old_file_contents = file.read()
    except:
        old_file_contents = ""
    new_file_contents = ""
    md5 = hashlib.md5()
    header = \
'''# AUTO-GENERATED FILE, DO NOT HAND EDIT!
# Created by %s
''' % thisInvocation
    new_file_contents += header
    for formatInfo in dictionary:
        s = json.dumps(formatInfo.to_dict())+"\n"
        #s = str(formatInfo.id) + ": " + formatInfo.format_str+", "+formatInfo.filename+", " + str(formatInfo.linenumber) + "\n"
        md5.update(s.encode('utf-8'))
        new_file_contents += s
    md5 = md5.hexdigest()
    new_file_contents += "# Dictionary md5 is %s\n" % (md5)

    # Only write the file if the contents changed.
    if new_file_contents != old_file_contents:
        with open(dictFilename,'w') as dictFile:
            dictFile.write(new_file_contents)
    
    # copy the dictionary file to the deploy directory
    if dictionaryDeployDir != None:
        deployedFile = "%s/%s.json" % (dictionaryDeployDir,md5)
        if os.path.isfile(deployedFile):
            if not theSameWithoutComments(dictFilename, deployedFile):
                raise ValueError("ERROR!  Files %s and %s exist but are not identical!!" % (dictFilename, deployedFile))
        shutil.copy2(dictFilename, deployedFile)


    # Read in the old dictionary header file so we can compare it to new
    # desired contents, to decide if we need to rewrite it.
    try:
        with open(headerFilename, 'r') as file:
            old_file_contents = file.read()
    except:
        old_file_contents = ""
    new_file_contents = ""

    lineEndings = os.linesep
    # for windows, override lineseperator, to force windows native line endings,
    # even if we're running cygwin with \n line endings.
    if sys.platform.startswith('win32') or sys.platform.startswith('cygwin'):
        lineEndings="\r\n"

    header = \
'''/* AUTO-GENERATED FILE, DO NOT HAND EDIT!
Created by %s
*/
''' % thisInvocation
    new_file_contents += header
    for formatInfo in dictionary:
        format_str_id_name = formatInfo.filename.replace("/","_").replace(".cpp","").replace(".c","").replace(".","_")+"_line_"+str(formatInfo.linenumber)
        new_file_contents += "#define %s %d\n" % (format_str_id_name, formatInfo.id)
        new_file_contents +="#define %s_PARAM_TYPE_BITMASK 0x%X\n" % (format_str_id_name, formatInfo.param_type_bitmask)
    md5array = "0x" + ", 0x".join([md5[i:i+2] for i in range(0, len(md5), 2)])
    new_file_contents +="#define FORMAT_STR_DICTIONARY_ID %s\n" % (md5array)

    # Only write the file if the contents changed.
    # This prevents changes to source code from regenerating the dictionary
    # file and changing it's timestamp unless it actually changed, so that
    # we won't needlessly recompile code that depends on it.
    if new_file_contents != old_file_contents:
        print("Contents changed, regenerating %s" % (headerFilename))
        with open(headerFilename,'w', newline=lineEndings) as hdrFile:
            hdrFile.write(new_file_contents)
    else:
        print("Contents haven't changed, not regenerating %s" % (headerFilename))

def format_specifier_list(format_str):
    return re.findall(r'%[0-9\.lh]*[cdeEfgGiuopsxX%]', format_str)

def param_type_bitmask(format_str):
    bitmask = 0
    format_specifiers = format_specifier_list(format_str)
    for format_specifier in reversed(format_specifiers):
        bitmask <<= 1
        if 'f' in format_specifier:
            bitmask |= 1
    return bitmask

inttypes_format_specifiers = {
    "PRId8": "d",
    "PRId16": "d",
    "PRId32": "d",
    "PRId64": "ld",
    "PRId64": "lld",
    "PRIi8": "i",
    "PRIi16": "i",
    "PRIi32": "i",
    "PRIi64": "li",
    "PRIi64": "lli",
    "PRIo8": "o",
    "PRIo16": "o",
    "PRIo32": "o",
    "PRIo64": "lo",
    "PRIo64": "llo",
    "PRIx8": "x",
    "PRIx16": "x",
    "PRIx32": "x",
    "PRIx64": "lx",
    "PRIx64": "llx",
    "PRIX8": "X",
    "PRIX16": "X",
    "PRIX32": "X",
    "PRIX64": "lX",
    "PRIX64": "llX",
    "PRIu8": "u",
    "PRIu16": "u",
    "PRIu32": "u",
    "PRIu64": "lu",
    "PRIu64": "llu"
}

class PrintfInfo:
    def __init__(self, id, format_str, filename, linenumber):
        self.id = id
        self.filename = filename
        self.linenumber = linenumber
        self.param_type_bitmask = param_type_bitmask(format_str)
        # replace inttypes format specifiers
        while '" PRI' in format_str:
            matchObj = re.search( r'.*" (.*) ".*', format_str)
            specifier = matchObj.group(1).strip()
            conversion = inttypes_format_specifiers[specifier]
            format_str = format_str.replace('" %s "' % specifier, conversion)
        self.format_str = format_str
    def to_dict(self):
        ret = {}
        ret['id'] = self.id
        ret['format'] = self.format_str
        ret['filename'] = self.filename
        ret['linenumber'] = self.linenumber
        ret['type_bitmask'] = self.param_type_bitmask
        return ret

def ProcessFile(inputFilename):
    try:
        escapedFilePath = inputFilename.replace("/","_").replace(".cpp","").replace(".c","").replace(".","_")
        with open(inputFilename, 'r') as inputFile:
            inputData = inputFile.read().splitlines() 
        global maxSubscriptions
        global format_str_id
        lineNumber = 0;
        for line in inputData:
            lineNumber+=1
            format_str = ""
            try:
                printStatement = ""
                if "#define ESCAPED_FILE_PATH" in line:
                    matchObj = re.search( r'#define\s*ESCAPED_FILE_PATH\s*(.*)', line)
                    escapedFilePathFromFile = matchObj.group(1).strip()
                    if escapedFilePath != escapedFilePathFromFile:
                        print(inputFilename + ":"+str(lineNumber) + ", " + escapedFilePathFromFile + " != " + escapedFilePath)
                        sedCmd = "sed -i '"+str(lineNumber)+"s/"+escapedFilePathFromFile+"/"+escapedFilePath+"/' "+inputFilename
                        #print("need to " + sedCmd)
                        os.system(sedCmd)
                elif "debugPrintf(" in line:
                    printStatement = "debugPrintf"
                    matchObj = re.search( r'debugPrintf\("(.*)"', line)
                    format_str = matchObj.group(1).strip()
                elif "debugWarn(" in line:
                    printStatement = "debugWarn"
                    matchObj = re.search( r'debugWarn\("(.*)"', line)
                    format_str = matchObj.group(1).strip()
                elif "debugError(" in line:
                    printStatement = "debugError"
                    matchObj = re.search( r'debugError\("(.*)"', line)
                    format_str = matchObj.group(1).strip()
                if format_str != "":
                    printInfo = PrintfInfo(format_str_id, format_str, inputFilename, lineNumber)
                    dictionary.append(printInfo)
                    format_str_id+=1
            except AttributeError:
                print("Regex search error on " + inputFilename + ", line " + str(lineNumber))
                print(line)
                sys.exit(1)
    except UnicodeDecodeError:
        print("UnicodeDecodeError on " + inputFilename) 

def main():
    if len(sys.argv) < 4:
        Usage();
    global thisInvocation
    thisInvocation = " ".join(sys.argv)
    inputDir = sys.argv[1]
    dictionaryFile = sys.argv[2]
    dictionaryHeaderFile = sys.argv[3]
    try:
        dictionaryDeployDir = sys.argv[4]
    except IndexError:
        dictionaryDeployDir = None

    global dictionary
    dictionary = []
    
    global format_str_id
    format_str_id = 0
    
    inputDirList = inputDir.split(',')
    for d in inputDirList:
        ProcessDir(d)

    try:
        os.makedirs(os.path.dirname(dictionaryFile))
    except:
        pass
    printDictionary(dictionaryFile, dictionaryHeaderFile, dictionaryDeployDir)

# main starts here
if __name__ == '__main__':
    main()



# classes to perform integer/float conversions.
import math
import numpy

class ResistanceConversion:
    @staticmethod
    def Convert(adc, v_ref, v_full, r1, r2, bit_depth):
        v_adc = adc * v_full / (2 ** bit_depth)
        if r2 == 0 and r1 != 0:
            r = r1 * ((v_ref / v_adc) - 1.0)
        elif r1 == 0 and r2 != 0:
            r = r2 * (v_adc / v_ref)/(1.0 - (v_adc/v_ref))
        else:
            r = (r1 * r2) / ((((v_ref / v_adc) - 1.0) * r1) - r2)
        return r

    @staticmethod
    def Invert(r, v_ref, v_full, r1, r2, bit_depth):
        if r2 == 0 and r1 != 0:
            v_adc = v_ref / ((r / r1) + 1.0)
        elif r2 != 0 and r1 == 0:
            v_adc = v_ref / ((r2 / r) + 1.0)
        else:
            v_adc = v_ref / (((r1 * r2) + (r2 * r))/(r1 * r) + 1.0)
        return int(v_adc * (2 ** bit_depth) / v_full)

class SteinhartHartConversion:
    @staticmethod
    def Convert(raw, a, b, c, d, r0, v_ref, v_full, r1, r2, bit_depth):
        r = ResistanceConversion.Convert(raw, v_ref, v_full, r1, r2, bit_depth)
        x = math.log(r / r0)
        temp_k = 1.0 / (a + b * x + c * (x ** 2) + d * (x ** 3))
        temp_c = temp_k - 273.15
        return temp_c

    @staticmethod
    def Invert(temp_c, a, b, c, d, r0, v_ref, v_full, r1, r2, bit_depth):
        temp_k = temp_c + 273.15
        roots = numpy.roots([d, c, b, a - (1.0 / temp_k)])
        if numpy.iscomplex(roots[-1]):
            raise # inversion did not work
        return ResistanceConversion.Invert(r0 * math.exp(roots[-1]),
                                           v_ref, v_full, r1, r2, bit_depth)

class CallendarVanDusenConversion:# Platinum RTD
    @staticmethod
    def Convert(raw, a, b, c, r0, v_ref, v_full, r1, r2, bit_depth):
        r = ResistanceConversion.Convert(raw, v_ref, v_full, r1, r2, bit_depth)
        if r > r0:
            return (-a + math.sqrt(a**2 - 4 * b * (1.0 - (r / r0)))) / (2 * b)
        t = ((r / r0) - 1.0) / (a + (100.0 * b))
        for i in range(3): # up to 3 iterations
            tn = t - ((1 + a * t + b * (t ** 2) + c * (t ** 3) * (t - 100) - (res / r0)) /
                      (a + 2 * b * t - 300 * c * (t**2) + 4 * c * (t**3)))
            if abs(tn - t) < 3.0: # arbitrary threshold to determine convergence
                return tn
            t = tn
        return t

    @staticmethod
    def Invert(temp_c, a, b, c, r0, v_ref, v_full, r1, r2, bit_depth):
        if temp_c > 0:
            r = r0 * (1 + a * temp_c + b * (temp_c ** 2))
        else:
            r = r0 * (1 + a * temp_c + b * (temp_c ** 2) + c * (temp_c ** 3) * (temp_c - 100))
        return ResistanceConversion.Invert(r, v_ref, v_full, r1, r2, bit_depth)



import msgtools.parser.parser as MsgParser
from msgtools.parser.MsgUtils import *

import msgtools.parser.cpp.language as language

language.namespace = "<MSGFULLNAME>_"
language.firstParam = "m_data"
language.firstParamDecl = "uint8_t* m_data"
language.const = ""
language.enumNamespace = 1
language.functionPrefix = "INLINE "
language.enumClass = ""

enums = language.enums
accessors = language.accessors
declarations = language.declarations
structPacking = language.structPacking
structUnpacking = language.structUnpacking
initCode = language.initCode

def fieldDefault(field):
    ret = field["Default"]
    if("Type" in field and "int" in field["Type"]) and ret > 2**31:
        ret = str(ret)+'u'
    return ret

def fieldInfo(field):
    ret = ""
    if "Default" in field:
        ret += '#define <MSGFULLNAME>_'+field["Name"]+'_DEFAULT ' + str(fieldDefault(field)) + "\n"
    if MsgParser.fieldCount(field) > 1:
        ret += '#define <MSGFULLNAME>_'+field["Name"]+'_COUNT ' + str(MsgParser.fieldCount(field)) + "\n"
    ret += '#define <MSGFULLNAME>_'+field["Name"]+'_OFFSET ' + str(MsgParser.fieldLocation(field)) + "\n"
    ret += '#define <MSGFULLNAME>_'+field["Name"]+'_SIZE ' + str(MsgParser.fieldSize(field)) + "\n"
    return ret

def fieldBitsInfo(field, bits, bitOffset, numBits):
    ret = ""
    if "Default" in bits:
        ret += '#define <MSGFULLNAME>_'+bits["Name"]+'_Default ' + str(bits["Default"]) + "\n"
    return ret

def fieldInfos(msg):
    ret = ""
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            ret += fieldInfo(field)
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    ret += fieldBitsInfo(field, bits, bitOffset, numBits)
                    bitOffset += numBits

    return ret

def getMsgID(msg):
    return baseGetMsgID("<MSGFULLNAME>_", "m_data", "CAST_ALL", 0, msg)
    
def setMsgID(msg):
    return baseSetMsgID("<MSGFULLNAME>_", "m_data, ", 0, 0, msg)



import can
import os
from PyQt5 import QtCore

class SocketcanCommands(QtCore.QObject):
    statusUpdate = QtCore.pyqtSignal(str)
    portsChanged = QtCore.pyqtSignal()
    def __init__(self, parent):
        super().__init__(parent)
        self._process = QtCore.QProcess(self)
        self._process.setProcessChannelMode(QtCore.QProcess.MergedChannels)
        self._process.readyReadStandardOutput.connect(self._printStdOut)
    
    # Adding a destructor does no good, it doesn't seem to get run (which is a known issue with Python).
    # Instead our stop() function below must be called when our parent's window closes.
    #def __del__(self):
    #    self.stop()

    def stop(self):
        if self._process.state() == QtCore.QProcess.Running:
            # closing the write channel seems to make the 'pkexec sh' exit.
            # trying to send ctrl-d to it, or calling terminate on the process, don't seem to work.
            self._process.closeWriteChannel()
            self._process.waitForFinished( 500 )
        if self._process.state() == QtCore.QProcess.Running:
            self.statusUpdate.emit("ERROR!  SocketcanCommands QProcess still running!")
            self._process.kill()

    def list(self):
        return can.detect_available_configs(['socketcan', "pcan"])
    
    def create(self, name):
        # Create the virtual CAN interface.
        self._run_command('ip link add dev %s type vcan' % name)
        self._run_command('ip link set %s up' % name)

    def configure(self, name, bitrate=1e6, data_bitrate=5e6):
        # configure
        self._run_command('ip link set %s down' % name)
        self._run_command('ip link set %s type can bitrate %d dbitrate %d fd on' % (name, bitrate, data_bitrate))
        self._run_command('ip link set %s up' % name)

    def delete(self, name):
        # bring down
        self._run_command('ip link del %s' % name)

    def _run_command(self, command):
        if self._process.state() == QtCore.QProcess.NotRunning:
            self._process.start('pkexec', ["sh"])
            self._process.waitForStarted()
        self._process.write(str.encode("echo %s\n%s\n" % (command,command), encoding='utf-8'))

    def _printStdOut(self):
        sender = self.sender()
        output = str(sender.readAllStandardOutput(), encoding='utf-8')
        # remove blank lines
        output = os.linesep.join([s for s in output.splitlines() if s])
        if output != "":
            self.statusUpdate.emit("socketcan: %s" % output)
        self.portsChanged.emit()




import json

# This class is used to load and save persistent configuration settings,
# from both files and devices.
# Specify a filename if you want to load settings from a file
# Specify a message namespace that contains the following messages, if you want to set/get settings in a device:
#   GetConfigKeys       - get list of keys stored in a device
#   SetConfigValue      - set the value of a key
#   GetConfigValue      - get the value of a key
#   DeleteConfigValue   - delete a key's value
#   CurrentConfigKeys   - the current list of keys stored in a device
#   CurrentConfigValue  - the current value of a key stored in a device
# Specify a connection if you want to talk to a device.
# Example:
#     cfg_settings = ConfigSettings("autopilot.cfg", M.Messages.Autopilot, cxn)
#     cfg_settings.SaveToDevice()
class ConfigSettings:
    def __init__(self, filename=None, msg_namespace=None, connection=None):
        self.settings = {}
        if filename:
            self.LoadFromFile(filename)
        self.msg_namespace = msg_namespace
        self.connection = connection

    def SaveToFile(self, filename):
        with open(filename, 'w') as f:
            for key, value in self.settings.items():
                d = {key: value}
                f.write(json.dumps(d)+"\n")

    def LoadFromFile(self, filename):
        self.settings = {}
        with open(filename, 'r') as f:
            for line in f:
                data = json.loads(line)
                for key, values in data.items():
                    self.settings[key] = values

    def SaveToDevice(self):
        for key, values in self.settings.items():
            try:
                # If the value is an empty string, delete the setting from the device.
                if values == "":
                    msg = self.msg_namespace.DeleteConfigValue()
                    msg.SetKey(key)
                else:
                    msg = self.msg_namespace.SetConfigValue()
                    msg.SetKey(key)
                    idx = 0
                    for value in values.split(","):
                        msg.SetValues(float(value), idx)
                        idx += 1
                    msg.SetCount(idx)
                self.connection.send(msg)
            except KeyError:
                print("ERROR! Invalid Key %s not in %s" % (key, self.items_by_key.keys()))

    def LoadFromDevice(self):
        MAX_RETRIES = 10
        RX_TIMEOUT = 1.0 # floating point seconds
        keys_to_request = {}

        # Do a number of retries, for timing out on receiving messages
        for i in range(MAX_RETRIES):
            # Request the list of config keys
            query_msg = self.msg_namespace.GetConfigKeys()
            self.connection.send(query_msg)
            # wait for a response
            msg = self.connection.recv(self.msg_namespace.CurrentConfigKeys, RX_TIMEOUT)

            # if we got a message with a list of config setting keys,
            # iterate through the keys and add them to the dictionary
            # of keys to request values for.
            if type(msg) == self.msg_namespace.CurrentConfigKeys:
                for i in range(msg.GetCount()):
                    keys_to_request[msg.GetKey(i)] = True
                break

        # Do a number of retries, for timing out on receiving messages
        for i in range(MAX_RETRIES):
            # Request values for all the keys we haven't received values for yet.
            for key in keys_to_request.keys():
                query_msg = self.msg_namespace.GetConfigValue()
                query_msg.SetKey(key)
                self.connection.send(query_msg)

            # Loop as long as we keep receiving messages without a timeout occurring
            while True:
                msg = self.connection.recv(self.msg_namespace.CurrentConfigValue, RX_TIMEOUT)
                if type(msg) == self.msg_namespace.CurrentConfigValue:
                    # if we got a message with the value of a config setting,
                    # store it in the dictionary, unless it's an invalid key
                    key_int = msg.GetKey(enumAsInt=True)
                    key = msg.GetKey()
                    if key == "Invalid" or key_int == 0:
                        return
                    if key in keys_to_request:
                        del keys_to_request[key]
                        count = msg.GetCount()
                        value = ""
                        for i in range(count):
                            value += str(msg.GetValues(i)) + ", "
                        value = value[:-2]
                        self.settings[key] = value
                        if len(keys_to_request) == 0:
                            break
                    else:
                        print("Key %s not in keys %s" % (key, list(keys_to_request.keys())))
                elif msg == None:
                    # If we timed out, then break out of the rx loop into the timeout loop.
                    break
                else:
                    print("Received unexpected message %s" % (msg))
        return self.settings
    CPP_FILE_TEMPLATE = '''\
// AUTOGENERATED FILE, DO NOT HAND EDIT!
// created by <CMDLINE> at <DATETIME>
#include "config_settings.h"
#include "<MSG_HEADER_NAME>"

int DefaultConfigSetting::Count() { return <SETTINGS_COUNT>; }
ConfigSettings::ConfigSetting<ConfigSettings::MAX_LEN>** DefaultConfigSetting::Defaults()
{
<CONFIG_SETTINGS>
    static ConfigSettings::ConfigSetting<ConfigSettings::MAX_LEN>* defaults[] = {
<CONFIG_DECLS>
    };
    return defaults;
}
'''
    def SaveToCppFile(self, filename, msg_header_name, enum_prefix):
        import datetime
        import sys
        if enum_prefix:
            enum_prefix = "(int)"+enum_prefix
        cfg_settings = ""
        for key, value in self.settings.items():
            value = value.replace(",", "f,")+"f"
            length = value.count(",")+1
            crc = 0
            cfg_settings += "    static const ConfigSettings::ConfigSetting<%s> %s_setting = {%s/*crc*/, %s, %s-1/*size=len-1*/, {%s}};\n" % (length, key, crc, enum_prefix+key, length, value)
        cfg_decls = ""
        if len(self.settings) > 0:
            for key, value in self.settings.items():
                length = value.count(",")+1
                cfg_decls += "        {(ConfigSettings::ConfigSetting<ConfigSettings::MAX_LEN>*)&%s_setting},\n" % (key)
        else:
            cfg_decls += "        {0},\n"
        cfg_decls = cfg_decls[:-2]
        output = ConfigSettings.CPP_FILE_TEMPLATE
        output = output.replace("<CMDLINE>", " ".join(sys.argv[:]))
        output = output.replace("<DATETIME>", str(datetime.datetime.now()))
        output = output.replace("<SETTINGS_COUNT>", str(len(self.settings)))
        output = output.replace("<CONFIG_SETTINGS>", cfg_settings)
        output = output.replace("<CONFIG_DECLS>", cfg_decls)
        output = output.replace("<MSG_HEADER_NAME>", msg_header_name)
        with open(filename, 'w') as f:
            f.write(output)



import msgtools.parser.parser as MsgParser
from msgtools.parser.MsgUtils import *

def outputSubdir(outDir, filename):
    return outDir + "/+" + filename

def fieldType(field):
    typeStr = field["Type"]
    if "int" in typeStr:
        return typeStr
    if typeStr == "float32":
        return "single";
    if typeStr == "float64":
        return "double";
    return "?"

def typeForScaledInt(field):
    numBits = MsgParser.fieldNumBits(field)
    if numBits > 24:
        return "double"
    return "single"

def set_field(msg, field):
    if fieldCount(field) == 1:
        ret = "<MSGFULLNAME>_Set%s(m_data, %s);" % (field["Name"], field["Name"])
    else:
        ret = '''\
for (int i=0; i < %d; i++)
{
    <MSGFULLNAME>_Set%s(m_data, %s[i], i);
}
''' % (fieldCount(field), field["Name"], field["Name"])
    return ret

def set_fields(msg):
    ret = []
    if "Fields" in msg:
        for field in msg["Fields"]:
            if "Bitfields" in field:
                for bitfield in field["Bitfields"]:
                    ret.append(set_field(msg, bitfield))
            else:
                ret.append(set_field(msg, field))

    return ret

def get_field(msg, field):
    if fieldCount(field) == 1:
        ret = "%s = <MSGFULLNAME>_Get%s(m_data);" % (field["Name"], field["Name"])
    else:
        ret = '''\
for (int i=0; i < %d; i++)
{
    %s[i] = <MSGFULLNAME>_Get%s(m_data, i);
}
''' % (fieldCount(field), field["Name"], field["Name"])
    return ret

def get_fields(msg):
    ret = []
    if "Fields" in msg:
        for field in msg["Fields"]:
            if "Bitfields" in field:
                for bitfield in field["Bitfields"]:
                    ret.append(get_field(msg, bitfield))
            else:
                ret.append(get_field(msg, field))

    return ret

def accessors(msg):
    return []

def enums(e):
    return ""

def declarations(msg, msg_enums):
    return []

def fieldDefault(field):
    return "0"

def initCode(msg):
    return []

def undefinedMsgId():
    return "-1"

def getMsgID(msg):
    ret = ""
    if "Fields" in msg:
        for field in msg["Fields"]:
            if "IDBits" in field:
                numBits = field["IDBits"]
                if "Enum" in field:
                    pass
                getStr = "obj."+matlabFieldName(msg,field)
                if "Enum" in field:
                    getStr += "AsInt"
                getStr = "uint32("+getStr+")"
                ret =  addShift(ret, getStr, numBits)
            if "Bitfields" in field:
                for bitfield in field["Bitfields"]:
                    if "IDBits" in bitfield:
                        numBits = bitfield["IDBits"]
                        if "Enum" in bitfield:
                            pass
                        getStr = "obj."+BitfieldName(field, bitfield)
                        if "Enum" in bitfield:
                            getStr += "AsInt"
                        getStr = "uint32("+getStr+")"
                        ret =  addShift(ret, getStr, numBits)
    return ret
    
def setMsgID(msg):
    ret = ""
    numBits = 0
    if "Fields" in msg:
        for field in reversed(msg["Fields"]):
            if "IDBits" in field:
                if numBits != 0:
                    ret += "\nid = bitshift(id, -" + str(numBits)+")\n"
                numBits = field["IDBits"]
                setStr = "bitand(id, "+Mask(numBits)+")"
                #if "Enum" in field:
                #    setStr = field["Enum"]+"("+setStr+")"
                ret +=  "obj."+matlabFieldName(msg,field)+" = "+setStr
            if "Bitfields" in field:
                for bitfield in reversed(field["Bitfields"]):
                    if "IDBits" in bitfield:
                        if numBits != 0:
                            ret += "\nid = id >> " + str(numBits)+"\n"
                        numBits = bitfield["IDBits"]
                        setStr = "bitand(id, "+Mask(numBits)+")"
                        #if "Enum" in bitfield:
                        #    setStr = bitfield["Enum"]+"("+setStr+")"
                        ret +=  "obj."+bitfield["Name"]+" = "+setStr
    return ret

oneOutputFilePerMsg = True



from PyQt5 import QtCore, QtGui, QtWidgets, QtNetwork, QtWebSockets
from PyQt5.QtCore import QObject
from PyQt5.QtWebSockets import QWebSocket
from PyQt5.QtWebSockets import QWebSocketServer

from msgtools.lib.messaging import Messaging

class WebSocketClientConnection(QObject):
    disconnected = QtCore.pyqtSignal(object)
    messagereceived = QtCore.pyqtSignal(object)

    def __init__(self, webSocket):
        super(WebSocketClientConnection, self).__init__(None)

        self.removeClient = QtWidgets.QPushButton("Remove")
        self.removeClient.pressed.connect(lambda: self.webSocket.close())
        self.statusLabel = QtWidgets.QLabel()
        self.subscriptions = {}
        self.subMask = ~0
        self.subValue = 0
        self.isHardwareLink = False

        self.webSocket = webSocket
        self.webSocket.binaryMessageReceived.connect(self.processBinaryMessage)
        self.webSocket.disconnected.connect(self.onDisconnected)

        self.name = "Web Client"
        self.hostLabel = QtWidgets.QLabel(self.webSocket.peerAddress().toString().replace("::ffff:",""))
        self.statusLabel.setText(self.name)

    def widget(self, index):
        if index == 0:
            return self.removeClient
        if index == 1:
            return self.statusLabel
        if index == 2:
            return self.hostLabel
        return None
            
    def processBinaryMessage(self, bytes):
        hdr = Messaging.hdr(bytes.data())
        self.messagereceived.emit(hdr)

    def onDisconnected(self):
        self.disconnected.emit(self)

    def sendMsg(self, msg):
        self.webSocket.sendBinaryMessage(msg.rawBuffer().raw)

class WebSocketServer(QObject):
    statusUpdate = QtCore.pyqtSignal(str)
    newConnection = QtCore.pyqtSignal(object)

    def __init__(self, portNumber):
        super(WebSocketServer, self).__init__(None)

        self.portNumber = portNumber
        self.webSocketServer = QWebSocketServer("MsgServer", QWebSocketServer.NonSecureMode)
        self.webSocketServer.newConnection.connect(self.onNewConnection)

    def start(self):
        if not self.webSocketServer.listen(QtNetwork.QHostAddress.Any, self.portNumber):
            self.statusUpdate.emit("Con't open WebSocket on port "+str(self.portNumber)+"!")

    def onNewConnection(self):
        connection = WebSocketClientConnection(self.webSocketServer.nextPendingConnection())
        self.newConnection.emit(connection)



# Based on code from https://github.com/WereCatf/pytimedinput, with the original author claiming this license:
'''MIT License

Copyright (c) 2020 WereCatf

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''
import sys
import time
from typing import Tuple, Union

if(sys.platform == "win32"):
    import msvcrt
    import ctypes
    from ctypes import wintypes
    def checkStdin():
        return msvcrt.kbhit()
    def readStdin():
        return msvcrt.getwch()
    def getStdoutSettings():
        savedConsoleSettings = wintypes.DWORD()
        kernel32 = ctypes.windll.kernel32
        # The Windows standard handle -11 is stdout
        kernel32.GetConsoleMode(kernel32.GetStdHandle(-11), ctypes.byref(savedConsoleSettings))
        return savedConsoleSettings
    def setStdoutSettings(savedConsoleSettings):
        kernel32 = ctypes.windll.kernel32
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), savedConsoleSettings)
    def enableStdoutAnsiEscape():
        kernel32 = ctypes.windll.kernel32
        # Enable ANSI escape sequence parsing
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)
else:
    import select
    import tty
    import termios
    def checkStdin():
        return select.select([sys.stdin], [], [], 0) == ([sys.stdin], [], [])
    def readStdin():
        return sys.stdin.read(1)
    def getStdoutSettings():
        return termios.tcgetattr(sys.stdin)
    def setStdoutSettings(savedConsoleSettings):
        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, savedConsoleSettings)
    def enableStdoutAnsiEscape():
        # Should be enabled by default under Linux (and OSX?), just set cbreak-mode
        tty.setcbreak(sys.stdin.fileno(), termios.TCSADRAIN)


def GetKey(timeout: int, allowCharacters: str = "") -> str:
    """Wait for the user to press a single key out of an optional list of allowed ones.

    Args:
        timeout (int, required): How many seconds to wait for input.
        allowCharacters (str, optional): Which characters the user is allowed to enter. Defaults to "", ie. any character.

    Returns:
        str: Which key the user pressed, None if the input timed out.
    """

    if(not sys.__stdin__.isatty()):
        raise RuntimeError("timedKey() requires an interactive shell, cannot continue.")

    savedConsoleSettings = getStdoutSettings()
    enableStdoutAnsiEscape()

    timeStart = time.time()

    user_input = None
    # Do a try/finally here, so we always restore savedConsoleSettings.
    # Otherwise the user terminal get messed up if the user does ctrl-c
    # to break, and characters that are typed do not appear after we exit.
    try:
        # This appears to busy wait!  Seems like it should do a sleep,
        # even if only briefly, and only if timeout is large!
        while(True):
            time_elapsed = time.time() - timeStart
            if(time_elapsed >= timeout):
                break
            time_remaining = timeout - time_elapsed
            if time_remaining > 0.010:
                time.sleep(min(time_remaining * 0.5, 0.1))
            if(checkStdin()):
                c = readStdin()
                if(len(allowCharacters) == 0 or c in allowCharacters):
                    user_input = c
                    break
    finally:
        setStdoutSettings(savedConsoleSettings)
    return user_input



#!/usr/bin/env python3
import os
import sys
import argparse
from datetime import datetime
from PyQt5 import QtGui, QtWidgets, QtCore

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)
from msgtools.lib.messaging import Messaging
import msgtools.lib.gui

DESCRIPTION='''
    Noisemaker is a utility for generating message traffic.  Messages are roughly 
    staggered with a constant delta between each message, triggered at a rate 33Hz rate.
    Python's built in timing is highly unstable.  Timing won't be extremeley precise.
'''

class NoiseMaker(msgtools.lib.gui.Gui):
    def __init__(self, parent=None):

        parser = argparse.ArgumentParser(description=DESCRIPTION)
        parser.add_argument('msgs', nargs='*', help='''
            One or more messages that you wish to send.  This entire list is sent each 
            time the message period is up.  Network messages are ignored.  By default we 
            send all known messages.''')
        parser = msgtools.lib.gui.Gui.addBaseArguments(parser)
        args = parser.parse_args()

        msgtools.lib.gui.Gui.__init__(self, "Noise Maker 0.1", args, parent)
        
        self.timeInfo = Messaging.findFieldInfo(Messaging.hdr.fields, "Time")
        
        # event-based way of getting messages
        self.RxMsg.connect(self.ProcessMessage)
        
        self.currentTime = 0
        
        self.msgTimer = QtCore.QTimer(self)
        self.msgTimer.setInterval(50)
        self.msgTimer.timeout.connect(self.msgTimeout)

        vbox = QtWidgets.QVBoxLayout()
        w = QtWidgets.QWidget(self)
        w.setLayout(vbox)
        self.setCentralWidget(w)
        
        warning = QtWidgets.QLabel(
        "THIS APP SPEWS RANDOM NOISE!\n"
        "Do not use this connected to real\n"
        "hardware or with a production system!")
        warning.setStyleSheet("QLabel { background-color : red; color : blue; }")
        warning.setAlignment(QtCore.Qt.AlignCenter)
        vbox.addWidget(warning)

        self.timeSlider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.timeSlider.setMinimum(1)
        self.timeSlider.setMaximum(5)
        vbox.addWidget(self.timeSlider)

        t = int(self.settings.value('time_slider', 10))
        self.timeSlider.setValue(t)
        self.timeSlider.valueChanged.emit(t)

        self.startStop = QtWidgets.QPushButton(self)
        self.startStop.setText('Start')
        self.startStop.clicked.connect(self.confirmNoise)
        vbox.addWidget(self.startStop)
        
        # check if user specified which messages we should output
        msgs = args.msgs
        
        # find a few messages to send at specified rates
        period = 0.1
        self.fieldNumber = 0
        self.msgPeriod = {}
        self.msgTxTime = {}
        for msgName in Messaging.MsgClassFromName:
            if not msgName.startswith("Network"):
                if len(msgs) > 0 and msgName not in msgs:
                    continue
                print("found message " + msgName)
                self.msgPeriod[msgName] = period
                period = period + 0.050
                if period > 0.5:
                    period = 0.1
                self.msgTxTime[msgName] = 0
    
    def confirmNoise(self):
        buttonReply = QtWidgets.QMessageBox.question(self,
            'NoiseMaker',
            "Are you sure you want to spew random data?\n"
            "This will have unpredictable/dangerous results!",
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.No)
        if buttonReply == QtWidgets.QMessageBox.Yes:
            self.startStop.clicked.disconnect(self.confirmNoise)
            self.startStop.clicked.connect(self.startStopFn)
            self.startStopFn()
        else:
            print('No clicked.')

    def startStopFn(self):
        if self.msgTimer.isActive():
            self.msgTimer.stop()
            self.startStop.setText('Start')
        else:
            self.msgTimer.start()
            self.startStop.setText('Stop')
    
    def msgTimeout(self):
        self.currentTime = datetime.now().timestamp()
        for msgName in self.msgTxTime:
            if self.currentTime > self.msgTxTime[msgName]:
                msgClass = Messaging.MsgClassFromName[msgName]
                self.msgTxTime[msgName] = self.currentTime + self.msgPeriod[msgName] * self.timeSlider.value()**2
                self.sendMsg(msgClass)
    
    def fieldValue(self, fieldInfo):
        minVal = 0
        maxVal = 255
        try:
            minVal = float(fieldInfo.minVal)
            maxVal = float(fieldInfo.maxVal)
        except (AttributeError, ValueError):
            minVal = 0
            maxVal = 255
        self.fieldNumber += 1
        if self.fieldNumber > 255:
            self.fieldNumber = 0
        # go from 10% below range to 10% above range
        bottom = minVal - 0.1 * (maxVal-minVal)
        range = 1.2 * (maxVal-minVal) 
        scale = self.fieldNumber / (255.0)
        return bottom + range * scale
                
    def sendMsg(self, msgClass):
        msg = msgClass()
        if self.timeInfo:
            hdr = Messaging.hdr(msg.rawBuffer())
            t = self.currentTime
            maxTime = self.timeInfo.maxVal
            # if the timestamp field is small, assume the user wants relative times since midnight.
            if maxTime != "DBL_MAX" and (maxTime == 'FLT_MAX' or float(maxTime) <= 2**32):
                t = (datetime.fromtimestamp(self.currentTime) - datetime.fromtimestamp(self.currentTime).replace(hour=0, minute=0, second=0, microsecond=0)).total_seconds()
            if self.timeInfo.units == "ms":
                t = t * 1000.0
            if self.timeInfo.type == "int":
                t = int(t)
            hdr.SetTime(t)

        for fieldInfo in msgClass.fields:
            if fieldInfo.units == 'ASCII':
                Messaging.set(msg, fieldInfo, 's'+str(self.fieldValue(fieldInfo)))
            else:
                if(fieldInfo.count == 1):
                    Messaging.set(msg, fieldInfo, str(self.fieldValue(fieldInfo)))
                    for bitInfo in fieldInfo.bitfieldInfo:
                        Messaging.set(msg, bitInfo, str(self.fieldValue(bitInfo)))
                else:
                    for i in range(0,fieldInfo.count):
                        Messaging.set(msg, fieldInfo, self.fieldValue(fieldInfo), i)
        self.SendMsg(msg)

    def ProcessMessage(self, msg):
        pass
    
    def on_close(self):
        self.settings.setValue('time_slider', self.timeSlider.value())

def main(args=None):
    app = QtWidgets.QApplication(sys.argv)
    msgApp = NoiseMaker()
    msgApp.show()
    sys.exit(app.exec_())

# main starts here
if __name__ == '__main__':
    main()
